# Licensed under a 3-clause BSD style license - see LICENSE.rst

from collections import OrderedDict
from textwrap import indent

from .coordinate_helpers import CoordinateHelper
from .coordinate_range import find_coordinate_range
from .frame import RectangularFrame, RectangularFrame1D


class CoordinatesMap:
    """
    A container for coordinate helpers that represents a coordinate system.

    This object can be used to access coordinate helpers by index (like a list)
    or by name (like a dictionary).

    Parameters
    ----------
    axes : :class:`~astropy.visualization.wcsaxes.WCSAxes`
        The axes the coordinate map belongs to.
    transform : `~matplotlib.transforms.Transform`, optional
        The transform for the data.
    coord_meta : dict, optional
        A dictionary providing additional metadata. This should include the keys
        ``type``, ``wrap``, and ``unit``. Each of these should be a list with as
        many items as the dimension of the coordinate system. The ``type``
        entries should be one of ``longitude``, ``latitude``, or ``scalar``, the
        ``wrap`` entries should give, for the longitude, the angle at which the
        coordinate wraps (and `None` otherwise), and the ``unit`` should give
        the unit of the coordinates as :class:`~astropy.units.Unit` instances.
        This can optionally also include a ``format_unit`` entry giving the
        units to use for the tick labels (if not specified, this defaults to
        ``unit``).
    frame_class : type, optional
        The class for the frame, which should be a subclass of
        :class:`~astropy.visualization.wcsaxes.frame.BaseFrame`. The default is to use a
        :class:`~astropy.visualization.wcsaxes.frame.RectangularFrame`
    previous_frame_path : `~matplotlib.path.Path`, optional
        When changing the WCS of the axes, the frame instance will change but
        we might want to keep re-using the same underlying matplotlib
        `~matplotlib.path.Path` - in that case, this can be passed to this
        keyword argument.
    """

    def __init__(
        self,
        axes,
        transform=None,
        coord_meta=None,
        frame_class=RectangularFrame,
        previous_frame_path=None,
    ):
        self._axes = axes
        self._transform = transform

        self.frame = frame_class(axes, self._transform, path=previous_frame_path)

        # Set up coordinates
        self._coords = []
        self._aliases = {}

        visible_count = 0

        for index in range(len(coord_meta["type"])):
            # Extract coordinate metadata
            coord_type = coord_meta["type"][index]
            coord_wrap = coord_meta["wrap"][index]
            coord_unit = coord_meta["unit"][index]
            name = coord_meta["name"][index]

            visible = True
            if "visible" in coord_meta:
                visible = coord_meta["visible"][index]

            format_unit = None
            if "format_unit" in coord_meta:
                format_unit = coord_meta["format_unit"][index]

            default_label = name[0] if isinstance(name, (tuple, list)) else name
            if "default_axis_label" in coord_meta:
                default_label = coord_meta["default_axis_label"][index]

            coord_index = None
            if visible:
                visible_count += 1
                coord_index = visible_count - 1

            self._coords.append(
                CoordinateHelper(
                    parent_axes=axes,
                    parent_map=self,
                    transform=self._transform,
                    coord_index=coord_index,
                    coord_type=coord_type,
                    coord_wrap=coord_wrap,
                    coord_unit=coord_unit,
                    format_unit=format_unit,
                    frame=self.frame,
                    default_label=default_label,
                )
            )

            # Set up aliases for coordinates
            if isinstance(name, tuple):
                for nm in name:
                    nm = nm.lower()
                    # Do not replace an alias already in the map if we have
                    # more than one alias for this axis.
                    if nm not in self._aliases:
                        self._aliases[nm] = index
            else:
                self._aliases[name.lower()] = index

    def __getitem__(self, item):
        if isinstance(item, str):
            return self._coords[self._aliases[item.lower()]]
        else:
            return self._coords[item]

    def __contains__(self, item):
        if isinstance(item, str):
            return item.lower() in self._aliases
        else:
            return 0 <= item < len(self._coords)

    def set_visible(self, visibility):
        raise NotImplementedError()

    def __iter__(self):
        yield from self._coords

    def grid(self, draw_grid=True, grid_type=None, **kwargs):
        """
        Plot gridlines for both coordinates.

        Standard matplotlib appearance options (color, alpha, etc.) can be
        passed as keyword arguments.

        Parameters
        ----------
        draw_grid : bool
            Whether to show the gridlines
        grid_type : { 'lines' | 'contours' }
            Whether to plot the contours by determining the grid lines in
            world coordinates and then plotting them in world coordinates
            (``'lines'``) or by determining the world coordinates at many
            positions in the image and then drawing contours
            (``'contours'``). The first is recommended for 2-d images, while
            for 3-d (or higher dimensional) cubes, the ``'contours'`` option
            is recommended. By default, 'lines' is used if the transform has
            an inverse, otherwise 'contours' is used.
        """
        for coord in self:
            coord.grid(draw_grid=draw_grid, grid_type=grid_type, **kwargs)

    def get_coord_range(self):
        xmin, xmax = self._axes.get_xlim()

        if isinstance(self.frame, RectangularFrame1D):
            extent = [xmin, xmax]
        else:
            ymin, ymax = self._axes.get_ylim()
            extent = [xmin, xmax, ymin, ymax]

        return find_coordinate_range(
            self._transform,
            extent,
            [coord.coord_type for coord in self if coord.coord_index is not None],
            [coord.coord_unit for coord in self if coord.coord_index is not None],
            [coord.coord_wrap for coord in self if coord.coord_index is not None],
        )

    def _as_table(self):
        # Import Table here to avoid importing the astropy.table package
        # every time astropy.visualization.wcsaxes is imported.
        from astropy.table import Table

        rows = []
        for icoord, coord in enumerate(self._coords):
            aliases = [key for key, value in self._aliases.items() if value == icoord]
            row = OrderedDict(
                [
                    ("index", icoord),
                    ("aliases", " ".join(aliases)),
                    ("type", coord.coord_type),
                    ("unit", coord.coord_unit),
                    ("wrap", coord.coord_wrap),
                    ("format_unit", coord.get_format_unit()),
                    ("visible", "no" if coord.coord_index is None else "yes"),
                ]
            )
            rows.append(row)
        return Table(rows=rows)

    def __repr__(self):
        s = f"<CoordinatesMap with {len(self._coords)} world coordinates:\n\n"
        table = indent(str(self._as_table()), "  ")
        return s + table + "\n\n>"



"""Main Lomb-Scargle Multiband Implementation"""

import numpy as np

from astropy import units as u
from astropy.time import Time, TimeDelta
from astropy.timeseries.binned import BinnedTimeSeries
from astropy.timeseries.periodograms.lombscargle import LombScargle
from astropy.timeseries.sampled import TimeSeries

from .implementations import available_methods, lombscargle_multiband
from .implementations.mle import construct_regularization, design_matrix, periodic_fit

__all__ = ["LombScargleMultiband"]


def has_units(obj):
    return hasattr(obj, "unit")


def get_unit(obj):
    return getattr(obj, "unit", 1)


def strip_units(*arrs):
    strip = lambda a: None if a is None else np.asarray(a)
    if len(arrs) == 1:
        return strip(arrs[0])
    else:
        return map(strip, arrs)


class LombScargleMultiband(LombScargle):
    """Compute the Lomb-Scargle Periodogram.

    This implementation is based on code presented in [1]_ and [2]_;
    if you use this functionality in an academic application, citation of
    those works would be appreciated.

    Parameters
    ----------
    t : array-like or `~astropy.units.Quantity` ['time']
        sequence of observation times
    y : array-like or `~astropy.units.Quantity`
        sequence of observations associated with times t
    bands : array-like
        sequence of passband labels associated with times t, each unique label
        defines a single band of data.
    dy : float, array-like, or `~astropy.units.Quantity`, optional
        error or sequence of observational errors associated with times t
    normalization : {'standard', 'model', 'log', 'psd'}, optional
        Normalization to use for the periodogram.
    nterms_base : int, optional
        number of frequency terms to use for the base model common to all bands.
        In the case of the fast algorithm, this parameter is passed along to
        the single band LombScargle method as the ``nterms`` parameter.
    nterms_band : int, optional
        number of frequency terms to use for the residuals between the base
        model and each individual band
    reg_base : float or None (default = None)
        amount of regularization to use on the base model parameters
    reg_band : float or None (default = 1E-6)
        amount of regularization to use on the band model parameters
    regularize_by_trace : bool (default = True)
        if True, then regularization is expressed in units of the trace of
        the normal matrix
    center_data : bool, optional
        if True, pre-center the data by subtracting the weighted mean
        of the input data.
    fit_mean : bool, optional
        if True, include a constant offset as part of the model at each
        frequency. This can lead to more accurate results, especially in the
        case of incomplete phase coverage. Only applicable to the "fast" method

    References
    ----------
    .. [1] Vanderplas, J., Connolly, A. Ivezic, Z. & Gray, A. *Introduction to
        astroML: Machine learning for astrophysics*. Proceedings of the
        Conference on Intelligent Data Understanding (2012)
    .. [2] VanderPlas, J. & Ivezic, Z. *Periodograms for Multiband Astronomical
        Time Series*. ApJ 812.1:18 (2015)
    """

    available_methods = available_methods()

    def __init__(
        self,
        t,
        y,
        bands,
        dy=None,
        normalization="standard",
        nterms_base=1,
        nterms_band=1,
        reg_base=None,
        reg_band=1e-6,
        regularize_by_trace=True,
        center_data=True,
        fit_mean=True,
    ):
        # If t is a TimeDelta, convert it to a quantity. The units we convert
        # to don't really matter since the user gets a Quantity back at the end
        # so can convert to any units they like.
        if isinstance(t, TimeDelta):
            t = t.to("day")

        # We want to expose self.t as being the times the user passed in, but
        # if the times are absolute, we need to convert them to relative times
        # internally, so we use self._trel and self._tstart for this.

        self.t = t

        if isinstance(self.t, Time):
            self._tstart = self.t[0]
            trel = (self.t - self._tstart).to(u.day)
        else:
            self._tstart = None
            trel = self.t

        self._trel, self.y, self.bands, self.dy = self._validate_inputs(
            trel, y, bands, dy
        )

        self.normalization = normalization
        self.nterms_base = nterms_base
        self.nterms_band = nterms_band
        self.reg_base = reg_base
        self.reg_band = reg_band
        self.regularize_by_trace = regularize_by_trace
        self.center_data = center_data
        self.fit_mean = fit_mean
        self.nterms = self.nterms_base  # determined by the base model params

    @classmethod
    def from_timeseries(
        cls,
        timeseries,
        signal_column=None,
        uncertainty_column=None,
        band_labels=None,
        **kwargs,
    ):
        """
        Initialize a multiband periodogram from a time series object.

        If a binned time series is passed, the time at the center of the bins is
        used. Also note that this method automatically gets rid of NaN/undefined
        values when initializing the periodogram.

        Parameters
        ----------
        signal_column : list
            The names of columns containing the signal values to use.
        uncertainty_column : list, optional
            The names of columns containing the errors on the signal.
        band_labels : list, optional
            The labels for each band, matched by index. If none, uses the
            labels of ``signal_column`` as band names.
        **kwargs
            Additional keyword arguments are passed to the initializer for this
            periodogram class.
        """
        if signal_column is None:
            raise ValueError(
                "signal_column_name should be set to a list of valid column names"
            )

        if band_labels is not None:
            if len(band_labels) != len(signal_column):
                raise ValueError(
                    "band_labels have an equal number of elements to signal_column"
                )
        else:
            band_labels = signal_column  # use the flux labels as band labels

        if isinstance(timeseries, TimeSeries):
            time = timeseries.time
        elif isinstance(timeseries, BinnedTimeSeries):
            time = timeseries.time_bin_center
        else:
            raise TypeError(
                "Input time series should be an instance of "
                "TimeSeries or BinnedTimeSeries"
            )

        # Build lombscargle_multiband inputs
        t = []
        y = []
        dy = []
        band = []
        for i, col in enumerate(signal_column):
            if np.ma.is_masked(timeseries[col]):
                signal_mask = ~timeseries[col].mask
            else:
                signal_mask = ~np.isnan(timeseries[col])

            if uncertainty_column is not None:
                dy_col = timeseries[uncertainty_column[i]]
                if np.ma.is_masked(dy_col):
                    signal_mask &= ~dy_col.mask
                else:
                    signal_mask &= ~np.isnan(dy_col)

            t.append(time[signal_mask].mjd * u.day)
            y.append(timeseries[col][signal_mask])
            band.append([band_labels[i]] * sum(signal_mask))
            dy.append(timeseries[uncertainty_column[i]][signal_mask])

        t = np.hstack(t)
        y = np.hstack(y)
        band = np.hstack(band)
        if uncertainty_column is not None:
            dy = np.hstack(dy)

        if len(dy) == 0:
            dy = None
        return cls(t, y, band, dy=dy, **kwargs)

    def _validate_inputs(self, t, y, bands, dy):
        # Validate shapes of inputs
        if dy is None:
            t, y, bands = np.broadcast_arrays(t, y, bands, subok=True)
        else:
            t, y, bands, dy = np.broadcast_arrays(t, y, bands, dy, subok=True)
        if t.ndim != 1:
            raise ValueError("Inputs (t, y, bands, dy) must be 1-dimensional")

        # validate units of inputs if any is a Quantity
        if any(has_units(arr) for arr in (t, y, bands, dy)):
            t, y = map(u.Quantity, (t, y))
            if dy is not None:
                dy = u.Quantity(dy)
                try:
                    dy = u.Quantity(dy, unit=y.unit)
                except u.UnitConversionError:
                    raise ValueError("Units of dy not equivalent " "to units of y")
        return t, y, bands, dy

    def autofrequency(
        self,
        samples_per_peak=5,
        nyquist_factor=5,
        minimum_frequency=None,
        maximum_frequency=None,
        return_freq_limits=False,
    ):
        """Determine a suitable frequency grid for data.
        Note that this assumes the peak width is driven by the observational
        baseline, which is generally a good assumption when the baseline is
        much larger than the oscillation period.
        If you are searching for periods longer than the baseline of your
        observations, this may not perform well.
        Even with a large baseline, be aware that the maximum frequency
        returned is based on the concept of "average Nyquist frequency", which
        may not be useful for irregularly-sampled data. The maximum frequency
        can be adjusted via the nyquist_factor argument, or through the
        maximum_frequency argument.

        Parameters
        ----------
        samples_per_peak : float, optional
            The approximate number of desired samples across the typical peak
        nyquist_factor : float, optional
            The multiple of the average nyquist frequency used to choose the
            maximum frequency if maximum_frequency is not provided.
        minimum_frequency : float, optional
            If specified, then use this minimum frequency rather than one
            chosen based on the size of the baseline.
        maximum_frequency : float, optional
            If specified, then use this maximum frequency rather than one
            chosen based on the average nyquist frequency.
        return_freq_limits : bool, optional
            if True, return only the frequency limits rather than the full
            frequency grid.

        Returns
        -------
        frequency : ndarray or `~astropy.units.Quantity` ['frequency']
            The heuristically-determined optimal frequency bin
        """
        if hasattr(self._trel, "unit"):
            unit = self._trel.unit
            trel = self._trel.to(u.day)  # do frequency calculations in days
        else:
            trel = self._trel
        baseline = trel.max() - trel.min()
        n_samples = trel.size

        df = 1.0 / baseline / samples_per_peak

        if minimum_frequency is None:
            minimum_frequency = 0.5 * df

        if maximum_frequency is None:
            avg_nyquist = 0.5 * n_samples / baseline
            maximum_frequency = nyquist_factor * avg_nyquist

        # Convert back to the input units
        if hasattr(self._trel, "unit"):
            df = df.to(1 / unit)
            minimum_frequency = minimum_frequency.to(1 / unit)
            maximum_frequency = maximum_frequency.to(1 / unit)

        n_freq = 1 + int(np.round((maximum_frequency - minimum_frequency) / df))

        if return_freq_limits:
            return minimum_frequency, minimum_frequency + df * (n_freq - 1)
        else:
            return minimum_frequency + df * np.arange(n_freq)

    def autopower(
        self,
        method="flexible",
        sb_method="auto",
        normalization="standard",
        samples_per_peak=5,
        nyquist_factor=5,
        minimum_frequency=None,
        maximum_frequency=None,
    ):
        """Compute Lomb-Scargle power at automatically-determined frequencies.

        Parameters
        ----------
        method : str, optional
            specify the multi-band lomb scargle implementation to use. Options are:

            - 'flexible': Constructs a common model, and an offset model per individual
                band. Applies regularization to the resulting model to constrain
                complexity.
            - 'fast': Passes each band individually through LombScargle (single-band),
                combines periodograms at the end by weight. Speed depends on single-band
                method chosen in 'sb_method'.

        sb_method : str, optional
            specify the single-band lomb scargle implementation to use, only in
            the case of using the 'fast' multiband method. Options are:

            - 'auto': choose the best method based on the input
            - 'fast': use the O[N log N] fast method. Note that this requires
              evenly-spaced frequencies: by default this will be checked unless
              ``assume_regular_frequency`` is set to True.
            - 'slow': use the O[N^2] pure-python implementation
            - 'cython': use the O[N^2] cython implementation. This is slightly
              faster than method='slow', but much more memory efficient.
            - 'chi2': use the O[N^2] chi2/linear-fitting implementation
            - 'fastchi2': use the O[N log N] chi2 implementation. Note that this
              requires evenly-spaced frequencies: by default this will be checked
              unless ``assume_regular_frequency`` is set to True.
            - 'scipy': use ``scipy.signal.lombscargle``, which is an O[N^2]
              implementation written in C. Note that this does not support
              heteroskedastic errors.

        normalization : {'standard', 'model', 'log', 'psd'}, optional
            If specified, override the normalization specified at instantiation.
        samples_per_peak : float, optional
            The approximate number of desired samples across the typical peak
        nyquist_factor : float, optional
            The multiple of the average nyquist frequency used to choose the
            maximum frequency if maximum_frequency is not provided.
        minimum_frequency : float or `~astropy.units.Quantity` ['frequency'], optional
            If specified, then use this minimum frequency rather than one
            chosen based on the size of the baseline. Should be `~astropy.units.Quantity`
            if inputs to LombScargle are `~astropy.units.Quantity`.
        maximum_frequency : float or `~astropy.units.Quantity` ['frequency'], optional
            If specified, then use this maximum frequency rather than one
            chosen based on the average nyquist frequency. Should be `~astropy.units.Quantity`
            if inputs to LombScargle are `~astropy.units.Quantity`.

        Returns
        -------
        frequency, power : ndarray
            The frequency and Lomb-Scargle power
        """
        frequency = self.autofrequency(
            samples_per_peak=samples_per_peak,
            nyquist_factor=nyquist_factor,
            minimum_frequency=minimum_frequency,
            maximum_frequency=maximum_frequency,
        )
        power = self.power(
            frequency, method=method, sb_method=sb_method, normalization=normalization
        )

        return frequency, power

    def power(
        self, frequency, method="flexible", sb_method="auto", normalization="standard"
    ):
        """Compute the Lomb-Scargle power at the given frequencies.

        Parameters
        ----------
        frequency : array-like or `~astropy.units.Quantity` ['frequency']
            frequencies (not angular frequencies) at which to evaluate the
            periodogram. Note that in order to use method='fast', frequencies
            must be regularly-spaced.
        method : str, optional
            specify the multi-band lomb scargle implementation to use. Options are:

            - 'flexible': Constructs a common model, and an offset model per individual
                band. Applies regularization to the resulting model to constrain
                complexity.
            - 'fast': Passes each band individually through LombScargle (single-band),
                combines periodograms at the end by weight. Speed depends on single-band
                method chosen in 'sb_method'.
        sb_method : str, optional
            specify the single-band lomb scargle implementation to use, only in
            the case of using the 'fast' multiband method. Options can be found
            in `~astropy.timeseries.LombScargle`.
        normalization : {'standard', 'model', 'log', 'psd'}, optional
            If specified, override the normalization specified at instantiation.

        Returns
        -------
        power : ndarray
            The Lomb-Scargle power at the specified frequency
        """
        if normalization is None:
            normalization = self.normalization
        frequency = self._validate_frequency(frequency)
        f_shape = np.shape(frequency)
        power = lombscargle_multiband(
            strip_units(self._trel),
            strip_units(self.y),
            strip_units(self.bands),
            dy=strip_units(self.dy),
            frequency=strip_units(np.ravel(frequency)),
            method=method,
            sb_method=sb_method,
            normalization=normalization,
            nterms_base=self.nterms_base,
            nterms_band=self.nterms_band,
            reg_base=self.reg_base,
            reg_band=self.reg_band,
            regularize_by_trace=self.regularize_by_trace,
            center_data=self.center_data,
            fit_mean=self.fit_mean,
        )
        return np.reshape(power * self._power_unit(normalization), f_shape)

    def design_matrix(self, frequency, t_fit=None, bands_fit=None):
        """Compute the design matrix for a given frequency

        Parameters
        ----------
        frequency : float
            the frequency for the model
        t_fit : array-like, `~astropy.units.Quantity`, or `~astropy.time.Time` (optional)
            Times (length ``n_samples``) at which to compute the model.
            If not specified, then the times and uncertainties of the input
            data are used.
        bands_fit : array-like, or str
            Bands to use in fitting, must be subset of bands in input data.

        Returns
        -------
        ndarray
            The design matrix for the model at the given frequency.
            This should have a shape of (``len(t)``, ``n_parameters``).

        See Also
        --------
        model
        model_parameters
        offset
        """
        if t_fit is None:
            t_fit, dy = strip_units(self._trel, self.dy)
        else:
            t_fit, dy = strip_units(
                self._validate_t(self._as_relative_time("t", t_fit)), None
            )

        if bands_fit is None:
            bands_fit = np.unique(self.bands)
        elif type(bands_fit) == str:
            bands_fit = [bands_fit]
        unique_bands = np.unique(bands_fit)
        unique_bands_fit = np.unique(bands_fit)

        bands_fit = bands_fit[:, np.newaxis]

        if not set(unique_bands_fit).issubset(set(unique_bands)):
            raise ValueError(
                "bands_fit does not match training data: "
                f"input: {set(unique_bands_fit)} output: {set(unique_bands)}"
            )

        t_fit, bands_fit = np.broadcast_arrays(t_fit, bands_fit)
        return design_matrix(
            t_fit.ravel(),
            bands_fit.ravel(),
            frequency,
            dy,
            nterms_base=self.nterms_base,
            nterms_band=self.nterms_band,
        )

    def model(self, t, frequency, bands_fit=None):
        """Compute the Lomb-Scargle model at the given frequency.

        The model at a particular frequency is a linear model:
        model = offset + dot(design_matrix, model_parameters)

        Parameters
        ----------
        t : array-like or `~astropy.units.Quantity` ['time']
            Times (length ``n_samples``) at which to compute the model.
        frequency : float
            the frequency for the model
        bands_fit : list or array-like
            the unique bands to fit in the model

        Returns
        -------
        y : np.ndarray
            The model fit corresponding to the input times.
            Will have shape (``n_bands``,``n_samples``).

        See Also
        --------
        design_matrix
        offset
        model_parameters
        """
        if bands_fit is None:
            bands_fit = np.unique(self.bands)

        frequency = self._validate_frequency(frequency)
        t = self._validate_t(self._as_relative_time("t", t))
        y_fit = periodic_fit(
            *strip_units(self._trel, self.y, self.dy),
            bands=self.bands,
            frequency=strip_units(frequency),
            t_fit=strip_units(t),
            bands_fit=bands_fit,
            center_data=self.center_data,
            nterms_base=self.nterms_base,
            nterms_band=self.nterms_band,
            reg_base=self.reg_base,
            reg_band=self.reg_band,
            regularize_by_trace=self.regularize_by_trace,
        )

        return y_fit * get_unit(self.y)

    def offset(self, t_fit=None, bands_fit=None):
        """Return the offset array of the model

        The offset array of the model is the (weighted) mean of the y values in each band.
        Note that if self.center_data is False, the offset is 0 by definition.

        Parameters
        ----------
        t_fit : array-like, `~astropy.units.Quantity`, or `~astropy.time.Time` (optional)
            Times (length ``n_samples``) at which to compute the model.
            If not specified, then the times and uncertainties of the input
            data are used.
        bands_fit : array-like, or str
            Bands to use in fitting, must be subset of bands in input data.

        Returns
        -------
        offset : array

        See Also
        --------
        design_matrix
        model
        model_parameters
        """
        if bands_fit is None:
            bands_fit = np.unique(self.bands)
        if t_fit is None:
            on_fit = False
            t_fit = self.t
        else:
            on_fit = True

        bands_fit = bands_fit[:, np.newaxis]
        unique_bands = np.unique(self.bands)
        unique_bands_fit = np.unique(bands_fit)

        if not set(unique_bands_fit).issubset(set(unique_bands)):
            raise ValueError(
                "bands_fit does not match training data: "
                f"input: {set(unique_bands_fit)} output: {set(unique_bands)}"
            )
        y, dy = strip_units(self.y, self.dy)

        if np.shape(t_fit) != np.shape(
            bands_fit
        ):  # No need to broadcast if bands map to timestamps
            t_fit, bands_fit = np.broadcast_arrays(t_fit, bands_fit)

        # need to make sure all unique filters are represented
        u, i = np.unique(
            np.concatenate([bands_fit.ravel(), unique_bands]), return_inverse=True
        )

        if not self.center_data:
            return 0

        if dy is None:
            dy = 1
        dy = np.broadcast_to(dy, y.shape)

        # Calculate ymeans -- per band
        ymeans = np.zeros(y.shape)  # filter specific means
        for band in unique_bands:
            mask = self.bands == band
            ymeans[mask] = np.average(y[mask], weights=1 / dy[mask] ** 2)
        ymeans_fit = ymeans[i[: -len(unique_bands)]]

        if on_fit:
            return ymeans_fit * get_unit(self.y)
        else:
            return ymeans * get_unit(self.y)

    def model_parameters(self, frequency, units=True):
        r"""Compute the best-fit model parameters at the given frequency.

        The model described by these parameters is:

        .. math::

            y(t; f, \vec{\theta}) = \theta_0 + \sum_{n=1}^{\tt nterms_base} [\theta_{2n-1}\sin(2\pi n f t) + \theta_{2n}\cos(2\pi n f t)]
            + \theta_0^{(k)} + \sum_{n=1}^{\tt nterms_band} [\theta_{2n-1}^{(k)}\sin(2\pi n f t) +

        where :math:`\vec{\theta}` is the array of parameters returned by this function.

        Parameters
        ----------
        frequency : float
            the frequency for the model
        units : bool
            If True (default), return design matrix with data units.

        Returns
        -------
        theta : np.ndarray (n_parameters,)
            The best-fit model parameters at the given frequency.

        See Also
        --------
        design_matrix
        model
        offset
        """
        frequency = self._validate_frequency(frequency)
        t, y, dy = strip_units(self._trel, self.y, self.dy)

        if self.center_data:
            y = y - strip_units(self.offset())

        dy = np.ones_like(y) if dy is None else np.asarray(dy)
        X = design_matrix(
            t,
            self.bands,
            frequency,
            dy=dy,
            nterms_base=self.nterms_base,
            nterms_band=self.nterms_band,
        )

        regularization = construct_regularization(
            self.bands,
            nterms_base=self.nterms_base,
            nterms_band=self.nterms_band,
            reg_base=self.reg_base,
            reg_band=self.reg_band,
        )

        M = np.dot(X.T, X)

        if regularization is not None:
            # M is being affected by operations on diag
            diag = M.ravel(order="K")[:: M.shape[0] + 1]

            if self.regularize_by_trace:
                diag += diag.sum() * np.asarray(regularization)
            else:
                diag += np.asarray(regularization)

        try:
            parameters = np.linalg.solve(M, np.dot(X.T, y / dy))
        except np.linalg.LinAlgError:
            parameters = np.dot(M, np.linalg.lstsq(X.T, y / dy)[0])
        if units:
            parameters = get_unit(self.y) * parameters

        return parameters

    def false_alarm_probability(self):
        """Not Implemented"""
        raise NotImplementedError

    def false_alarm_level(self):
        """Not Implemented"""
        raise NotImplementedError

    def distribution(self):
        """Not Implemented"""
        raise NotImplementedError



# Licensed under a 3-clause BSD style license - see LICENSE.rst

# STDLIB
import inspect
import random

# THIRD PARTY
import numpy as np
import pytest

# LOCAL
from astropy.cosmology.core import Cosmology
from astropy.cosmology.io.model import _CosmologyModel, from_model, to_model
from astropy.cosmology.tests.helper import get_redshift_methods
from astropy.modeling.models import Gaussian1D
from astropy.utils.compat.optional_deps import HAS_SCIPY

from .base import ToFromDirectTestBase, ToFromTestMixinBase

###############################################################################


class ToFromModelTestMixin(ToFromTestMixinBase):
    """Tests for a Cosmology[To/From]Format with ``format="astropy.model"``.

    This class will not be directly called by :mod:`pytest` since its name does
    not begin with ``Test``. To activate the contained tests this class must
    be inherited in a subclass. Subclasses must define a :func:`pytest.fixture`
    ``cosmo`` that returns/yields an instance of a |Cosmology|.
    See ``TestCosmologyToFromFormat`` or ``TestCosmology`` for examples.
    """

    @pytest.fixture(scope="class")
    def method_name(self, cosmo):
        # get methods, ignoring private and dunder
        methods = get_redshift_methods(cosmo, include_private=False, include_z2=True)

        # dynamically detect ABC and optional dependencies
        for n in tuple(methods):
            params = inspect.signature(getattr(cosmo, n)).parameters.keys()

            ERROR_SEIVE = (NotImplementedError, ValueError)
            #              # ABC                can't introspect for good input
            if not HAS_SCIPY:
                ERROR_SEIVE = ERROR_SEIVE + (ModuleNotFoundError,)

            args = np.arange(len(params)) + 1
            try:
                getattr(cosmo, n)(*args)
            except ERROR_SEIVE:
                methods.discard(n)

        # TODO! pytest doesn't currently allow multiple yields (`cosmo`) so
        # testing with 1 random method
        # yield from methods
        return random.choice(tuple(methods)) if methods else None

    # ===============================================================

    def test_fromformat_model_wrong_cls(self, from_format):
        """Test when Model is not the correct class."""
        model = Gaussian1D(amplitude=10, mean=14)

        with pytest.raises(AttributeError):
            from_format(model)

    def test_toformat_model_not_method(self, to_format):
        """Test when method is not a method."""
        with pytest.raises(AttributeError):
            to_format("astropy.model", method="this is definitely not a method.")

    def test_toformat_model_not_callable(self, to_format):
        """Test when method is actually an attribute."""
        with pytest.raises(ValueError):
            to_format("astropy.model", method="name")

    def test_toformat_model(self, cosmo, to_format, method_name):
        """Test cosmology -> astropy.model."""
        if method_name is None:  # no test if no method
            return

        model = to_format("astropy.model", method=method_name)
        assert isinstance(model, _CosmologyModel)

        # Parameters
        expect = tuple(n for n in cosmo.__parameters__ if getattr(cosmo, n) is not None)
        assert model.param_names == expect

        # scalar result
        args = np.arange(model.n_inputs) + 1

        got = model.evaluate(*args)
        expected = getattr(cosmo, method_name)(*args)
        assert np.all(got == expected)

        got = model(*args)
        expected = getattr(cosmo, method_name)(*args)
        np.testing.assert_allclose(got, expected)

        # vector result
        if "scalar" not in method_name:
            args = (np.ones((model.n_inputs, 3)).T + np.arange(model.n_inputs)).T

            got = model.evaluate(*args)
            expected = getattr(cosmo, method_name)(*args)
            assert np.all(got == expected)

            got = model(*args)
            expected = getattr(cosmo, method_name)(*args)
            np.testing.assert_allclose(got, expected)

    def test_tofromformat_model_instance(
        self, cosmo_cls, cosmo, method_name, to_format, from_format
    ):
        """Test cosmology -> astropy.model -> cosmology."""
        if method_name is None:  # no test if no method
            return

        # ------------
        # To Model
        # this also serves as a test of all added methods / attributes
        # in _CosmologyModel.

        model = to_format("astropy.model", method=method_name)

        assert isinstance(model, _CosmologyModel)
        assert model.cosmology_class is cosmo_cls
        assert model.cosmology == cosmo
        assert model.method_name == method_name

        # ------------
        # From Model

        # it won't error if everything matches up
        got = from_format(model, format="astropy.model")
        assert got == cosmo
        assert set(cosmo.meta.keys()).issubset(got.meta.keys())
        # Note: model adds parameter attributes to the metadata

        # also it auto-identifies 'format'
        got = from_format(model)
        assert got == cosmo
        assert set(cosmo.meta.keys()).issubset(got.meta.keys())

    def test_fromformat_model_subclass_partial_info(self):
        """
        Test writing from an instance and reading from that class.
        This works with missing information.
        """
        pass  # there's no partial information with a Model

    @pytest.mark.parametrize("format", [True, False, None, "astropy.model"])
    def test_is_equivalent_to_model(self, cosmo, method_name, to_format, format):
        """Test :meth:`astropy.cosmology.Cosmology.is_equivalent`.

        This test checks that Cosmology equivalency can be extended to any
        Python object that can be converted to a Cosmology -- in this case
        a model.
        """
        if method_name is None:  # no test if no method
            return

        obj = to_format("astropy.model", method=method_name)
        assert not isinstance(obj, Cosmology)

        is_equiv = cosmo.is_equivalent(obj, format=format)
        assert is_equiv is (format is not False)


class TestToFromModel(ToFromDirectTestBase, ToFromModelTestMixin):
    """Directly test ``to/from_model``."""

    def setup_class(self):
        self.functions = {"to": to_model, "from": from_model}



# Licensed under a 3-clause BSD style license - see LICENSE.rst

import pytest
from numpy.testing import assert_allclose, assert_equal

from astropy import units as u
from astropy.tests.helper import assert_quantity_allclose
from astropy.time import Time, TimeDelta
from astropy.timeseries.binned import BinnedTimeSeries
from astropy.timeseries.periodograms import BoxLeastSquares, LombScargle
from astropy.utils.data import get_pkg_data_filename

CSV_FILE = get_pkg_data_filename("data/binned.csv")


def test_empty_initialization():
    ts = BinnedTimeSeries()
    ts["time_bin_start"] = Time([1, 2, 3], format="mjd")


def test_empty_initialization_invalid():
    # Make sure things crash when the first column added is not a time column

    ts = BinnedTimeSeries()
    with pytest.raises(
        ValueError,
        match=(
            r"BinnedTimeSeries object is invalid - expected 'time_bin_start' as the"
            r" first column but found 'flux'"
        ),
    ):
        ts["flux"] = [1, 2, 3]


def test_initialization_time_bin_invalid():
    # Make sure things crash when time_bin_* is passed incorrectly.

    with pytest.raises(TypeError, match=r"'time_bin_start' has not been specified"):
        BinnedTimeSeries(data=[[1, 4, 3]])

    with pytest.raises(
        TypeError, match=r"Either 'time_bin_size' or 'time_bin_end' should be specified"
    ):
        BinnedTimeSeries(time_bin_start="2016-03-22T12:30:31", data=[[1, 4, 3]])


def test_initialization_time_bin_both():
    # Make sure things crash when time_bin_* is passed twice.
    MESSAGE = r"'{}' has been given both in the table and as a keyword argument"

    with pytest.raises(TypeError, match=MESSAGE.format("time_bin_start")):
        BinnedTimeSeries(
            data={"time_bin_start": ["2016-03-22T12:30:31"]},
            time_bin_start="2016-03-22T12:30:31",
        )

    with pytest.raises(TypeError, match=MESSAGE.format("time_bin_size")):
        BinnedTimeSeries(
            data={"time_bin_size": ["2016-03-22T12:30:31"]}, time_bin_size=[1] * u.s
        )


def test_initialization_time_bin_size():
    # Make sure things crash when time_bin_size has no units

    with pytest.raises(
        TypeError, match=r"'time_bin_size' should be a Quantity or a TimeDelta"
    ):
        BinnedTimeSeries(
            data={"time": ["2016-03-22T12:30:31"]},
            time_bin_start="2016-03-22T12:30:31",
            time_bin_size=1,
        )

    # TimeDelta for time_bin_size
    ts = BinnedTimeSeries(
        data={"time": ["2016-03-22T12:30:31"]},
        time_bin_start="2016-03-22T12:30:31",
        time_bin_size=TimeDelta(1, format="jd"),
    )
    assert isinstance(ts.time_bin_size, u.quantity.Quantity)


def test_initialization_time_bin_start_scalar():
    # Make sure things crash when time_bin_start is a scalar with no time_bin_size

    with pytest.raises(
        TypeError, match=r"'time_bin_start' is scalar, so 'time_bin_size' is required"
    ):
        BinnedTimeSeries(
            data={"time": ["2016-03-22T12:30:31"]},
            time_bin_start=Time(1, format="mjd"),
            time_bin_end=Time(1, format="mjd"),
        )


def test_initialization_n_bins_invalid_arguments():
    # Make sure an exception is raised when n_bins is passed as an argument while
    # any of the parameters 'time_bin_start' or 'time_bin_end' is not scalar.

    with pytest.raises(
        TypeError,
        match=(
            r"'n_bins' cannot be specified if 'time_bin_start' or 'time_bin_size' are"
            r" not scalar'"
        ),
    ):
        BinnedTimeSeries(
            time_bin_start=Time([1, 2, 3], format="cxcsec"),
            time_bin_size=1 * u.s,
            n_bins=10,
        )


def test_initialization_n_bins():
    # Make sure things crash with incorrect n_bins

    with pytest.raises(
        TypeError,
        match=(
            r"'n_bins' has been given and it is not the same length as the input data\."
        ),
    ):
        BinnedTimeSeries(
            data={"time": ["2016-03-22T12:30:31"]},
            time_bin_start=Time(1, format="mjd"),
            time_bin_size=1 * u.s,
            time_bin_end=Time(1, format="mjd"),
            n_bins=10,
        )


def test_initialization_non_scalar_time():
    # Make sure things crash with incorrect size of time_bin_start

    with pytest.raises(
        ValueError,
        match=r"Length of 'time_bin_start' \(2\) should match table length \(1\)",
    ):
        BinnedTimeSeries(
            data={"time": ["2016-03-22T12:30:31"]},
            time_bin_start=["2016-03-22T12:30:31", "2016-03-22T12:30:32"],
            time_bin_size=1 * u.s,
            time_bin_end=Time(1, format="mjd"),
        )

    with pytest.raises(
        TypeError, match=r"Either 'time_bin_size' or 'time_bin_end' should be specified"
    ):
        BinnedTimeSeries(
            data={"time": ["2016-03-22T12:30:31"]},
            time_bin_start=["2016-03-22T12:30:31"],
            time_bin_size=None,
            time_bin_end=None,
        )


def test_even_contiguous():
    # Initialize a ``BinnedTimeSeries`` with even contiguous bins by specifying
    # the bin width:

    ts = BinnedTimeSeries(
        time_bin_start="2016-03-22T12:30:31", time_bin_size=3 * u.s, data=[[1, 4, 3]]
    )

    assert_equal(
        ts.time_bin_start.isot,
        [
            "2016-03-22T12:30:31.000",
            "2016-03-22T12:30:34.000",
            "2016-03-22T12:30:37.000",
        ],
    )

    assert_equal(
        ts.time_bin_center.isot,
        [
            "2016-03-22T12:30:32.500",
            "2016-03-22T12:30:35.500",
            "2016-03-22T12:30:38.500",
        ],
    )

    assert_equal(
        ts.time_bin_end.isot,
        [
            "2016-03-22T12:30:34.000",
            "2016-03-22T12:30:37.000",
            "2016-03-22T12:30:40.000",
        ],
    )


def test_uneven_contiguous():
    # Initialize a ``BinnedTimeSeries`` with uneven contiguous bins by giving an
    # end time:

    ts = BinnedTimeSeries(
        time_bin_start=[
            "2016-03-22T12:30:31",
            "2016-03-22T12:30:32",
            "2016-03-22T12:30:40",
        ],
        time_bin_end="2016-03-22T12:30:55",
        data=[[1, 4, 3]],
    )

    assert_equal(
        ts.time_bin_start.isot,
        [
            "2016-03-22T12:30:31.000",
            "2016-03-22T12:30:32.000",
            "2016-03-22T12:30:40.000",
        ],
    )

    assert_equal(
        ts.time_bin_center.isot,
        [
            "2016-03-22T12:30:31.500",
            "2016-03-22T12:30:36.000",
            "2016-03-22T12:30:47.500",
        ],
    )

    assert_equal(
        ts.time_bin_end.isot,
        [
            "2016-03-22T12:30:32.000",
            "2016-03-22T12:30:40.000",
            "2016-03-22T12:30:55.000",
        ],
    )


def test_uneven_non_contiguous():
    # Initialize a ``BinnedTimeSeries`` with uneven non-contiguous bins with
    # lists of start times, bin sizes and data:

    ts = BinnedTimeSeries(
        time_bin_start=[
            "2016-03-22T12:30:31",
            "2016-03-22T12:30:38",
            "2016-03-22T12:34:40",
        ],
        time_bin_size=[5, 100, 2] * u.s,
        data=[[1, 4, 3]],
    )

    assert_equal(
        ts.time_bin_start.isot,
        [
            "2016-03-22T12:30:31.000",
            "2016-03-22T12:30:38.000",
            "2016-03-22T12:34:40.000",
        ],
    )

    assert_equal(
        ts.time_bin_center.isot,
        [
            "2016-03-22T12:30:33.500",
            "2016-03-22T12:31:28.000",
            "2016-03-22T12:34:41.000",
        ],
    )

    assert_equal(
        ts.time_bin_end.isot,
        [
            "2016-03-22T12:30:36.000",
            "2016-03-22T12:32:18.000",
            "2016-03-22T12:34:42.000",
        ],
    )


def test_uneven_non_contiguous_full():
    # Initialize a ``BinnedTimeSeries`` with uneven non-contiguous bins by
    # specifying the start and end times for the bins:

    ts = BinnedTimeSeries(
        time_bin_start=[
            "2016-03-22T12:30:31",
            "2016-03-22T12:30:33",
            "2016-03-22T12:30:40",
        ],
        time_bin_end=[
            "2016-03-22T12:30:32",
            "2016-03-22T12:30:35",
            "2016-03-22T12:30:41",
        ],
        data=[[1, 4, 3]],
    )

    assert_equal(
        ts.time_bin_start.isot,
        [
            "2016-03-22T12:30:31.000",
            "2016-03-22T12:30:33.000",
            "2016-03-22T12:30:40.000",
        ],
    )

    assert_equal(
        ts.time_bin_center.isot,
        [
            "2016-03-22T12:30:31.500",
            "2016-03-22T12:30:34.000",
            "2016-03-22T12:30:40.500",
        ],
    )

    assert_equal(
        ts.time_bin_end.isot,
        [
            "2016-03-22T12:30:32.000",
            "2016-03-22T12:30:35.000",
            "2016-03-22T12:30:41.000",
        ],
    )


def test_read_empty():
    with pytest.raises(
        ValueError,
        match=(
            r"``time_bin_start_column`` should be provided since the default Table"
            r" readers are being used\."
        ),
    ):
        BinnedTimeSeries.read(CSV_FILE, format="csv")


def test_read_no_size_end():
    with pytest.raises(
        ValueError,
        match=(
            r"Either `time_bin_end_column` or `time_bin_size_column` should be"
            r" provided\."
        ),
    ):
        BinnedTimeSeries.read(
            CSV_FILE, time_bin_start_column="time_start", format="csv"
        )


def test_read_both_extra_bins():
    with pytest.raises(
        ValueError,
        match=r"Cannot specify both `time_bin_end_column` and `time_bin_size_column`\.",
    ):
        BinnedTimeSeries.read(
            CSV_FILE,
            time_bin_start_column="time_start",
            time_bin_end_column="END",
            time_bin_size_column="bin_size",
            format="csv",
        )


def test_read_size_no_unit():
    with pytest.raises(
        ValueError,
        match=(
            r"The bin size unit should be specified as an astropy Unit using"
            r" ``time_bin_size_unit``\."
        ),
    ):
        BinnedTimeSeries.read(
            CSV_FILE,
            time_bin_start_column="time_start",
            time_bin_size_column="bin_size",
            format="csv",
        )


def test_read_start_time_missing():
    with pytest.raises(
        ValueError, match=r"Bin start time column 'abc' not found in the input data\."
    ):
        BinnedTimeSeries.read(
            CSV_FILE,
            time_bin_start_column="abc",
            time_bin_size_column="bin_size",
            time_bin_size_unit=u.second,
            format="csv",
        )


def test_read_end_time_missing():
    with pytest.raises(
        ValueError, match=r"Bin end time column 'missing' not found in the input data\."
    ):
        BinnedTimeSeries.read(
            CSV_FILE,
            time_bin_start_column="time_start",
            time_bin_end_column="missing",
            format="csv",
        )


def test_read_size_missing():
    with pytest.raises(
        ValueError, match=r"Bin size column 'missing' not found in the input data\."
    ):
        BinnedTimeSeries.read(
            CSV_FILE,
            time_bin_start_column="time_start",
            time_bin_size_column="missing",
            time_bin_size_unit=u.second,
            format="csv",
        )


def test_read_time_unit_missing():
    with pytest.raises(
        ValueError,
        match=(
            r"The bin size unit should be specified as an astropy Unit using"
            r" ``time_bin_size_unit``\."
        ),
    ):
        BinnedTimeSeries.read(
            CSV_FILE,
            time_bin_start_column="time_start",
            time_bin_size_column="bin_size",
            format="csv",
        )


def test_read():
    timeseries = BinnedTimeSeries.read(
        CSV_FILE,
        time_bin_start_column="time_start",
        time_bin_end_column="time_end",
        format="csv",
    )
    assert timeseries.colnames == [
        "time_bin_start",
        "time_bin_size",
        "bin_size",
        "A",
        "B",
        "C",
        "D",
        "E",
        "F",
    ]
    assert len(timeseries) == 10
    assert timeseries["B"].sum() == 1151.54

    timeseries = BinnedTimeSeries.read(
        CSV_FILE,
        time_bin_start_column="time_start",
        time_bin_size_column="bin_size",
        time_bin_size_unit=u.second,
        format="csv",
    )
    assert timeseries.colnames == [
        "time_bin_start",
        "time_bin_size",
        "time_end",
        "A",
        "B",
        "C",
        "D",
        "E",
        "F",
    ]
    assert len(timeseries) == 10
    assert timeseries["B"].sum() == 1151.54


@pytest.mark.parametrize("cls", [BoxLeastSquares, LombScargle])
def test_periodogram(cls):
    # Note that we don't need to check the actual results from the periodogram
    # classes here since these are tested extensively in
    # astropy.timeseries.periodograms.

    ts = BinnedTimeSeries(
        time_bin_start="2016-03-22T12:30:31",
        time_bin_size=3 * u.s,
        data=[[1, 4, 3], [3, 4, 3]],
        names=["a", "b"],
    )

    p1 = cls.from_timeseries(ts, "a")
    assert isinstance(p1, cls)
    assert_allclose(p1.t.jd, ts.time_bin_center.jd)
    assert_equal(p1.y, ts["a"])
    assert p1.dy is None

    p2 = cls.from_timeseries(ts, "a", uncertainty="b")
    assert_quantity_allclose(p2.dy, ts["b"])

    p3 = cls.from_timeseries(ts, "a", uncertainty=0.1)
    assert_allclose(p3.dy, 0.1)



# Licensed under a 3-clause BSD style license

import os

from setuptools import Extension

SRC_DIR = os.path.join(os.path.dirname(__file__), "src")


def get_extensions():
    return [
        Extension(
            "astropy.io.fits._tiled_compression._compression",
            sources=[
                os.path.join(SRC_DIR, "compression.c"),
                os.path.join(SRC_DIR, "unquantize.c"),
                os.path.join("cextern", "cfitsio", "lib", "pliocomp.c"),
                os.path.join("cextern", "cfitsio", "lib", "ricecomp.c"),
                os.path.join("cextern", "cfitsio", "lib", "fits_hcompress.c"),
                os.path.join("cextern", "cfitsio", "lib", "fits_hdecompress.c"),
                os.path.join("cextern", "cfitsio", "lib", "quantize.c"),
            ],
            include_dirs=[SRC_DIR],
        )
    ]



# Licensed under a 3-clause BSD style license - see LICENSE.rst

import os

import numpy as np
import pytest

from astropy.io.misc.hdf5 import meta_path
from astropy.table import Column, QTable, Table
from astropy.table.table_helpers import simple_table
from astropy.units import allclose as quantity_allclose
from astropy.units.quantity import QuantityInfo
from astropy.utils.compat import NUMPY_LT_1_22
from astropy.utils.compat.optional_deps import HAS_H5PY
from astropy.utils.data import get_pkg_data_filename
from astropy.utils.exceptions import AstropyUserWarning
from astropy.utils.misc import _NOT_OVERWRITING_MSG_MATCH

if HAS_H5PY:
    import h5py

from astropy.io.tests.mixin_columns import compare_attrs, mixin_cols, serialized_names

# HDF5 does not support object dtype (since it stores binary representations).
unsupported_cols = {
    name: col
    for name, col in mixin_cols.items()
    if (isinstance(col, np.ndarray) and col.dtype.kind == "O")
}
mixin_cols = {
    name: col for name, col in mixin_cols.items() if name not in unsupported_cols
}

ALL_DTYPES = [
    np.uint8,
    np.uint16,
    np.uint32,
    np.uint64,
    np.int8,
    np.int16,
    np.int32,
    np.int64,
    np.float32,
    np.float64,
    np.bool_,
    "|S3",
]


def _default_values(dtype):
    if dtype == np.bool_:
        return [0, 1, 1]
    elif dtype == "|S3":
        return [b"abc", b"def", b"ghi"]
    else:
        return [1, 2, 3]


@pytest.fixture
def home_is_tmpdir(monkeypatch, tmp_path):
    """
    Pytest fixture to run a test case with tilde-prefixed paths.

    In the tilde-path case, environment variables are temporarily
    modified so that '~' resolves to the temp directory.
    """
    # For Unix
    monkeypatch.setenv("HOME", str(tmp_path))
    # For Windows
    monkeypatch.setenv("USERPROFILE", str(tmp_path))


@pytest.mark.skipif(not HAS_H5PY, reason="requires h5py")
def test_write_nopath(tmp_path):
    test_file = tmp_path / "test.hdf5"
    t1 = Table()
    t1.add_column(Column(name="a", data=[1, 2, 3]))

    with pytest.warns(
        UserWarning, match="table path was not set via the path= argument"
    ):
        t1.write(test_file)

    t1 = Table.read(test_file, path="__astropy_table__")


@pytest.mark.skipif(not HAS_H5PY, reason="requires h5py")
def test_write_nopath_nonempty(tmp_path):
    test_file = tmp_path / "test.hdf5"
    t1 = Table()
    t1.add_column(Column(name="a", data=[1, 2, 3]))

    t1.write(test_file, path="bubu")

    with pytest.raises(ValueError) as exc:
        t1.write(test_file, append=True)

    assert "table path should always be set via the path=" in exc.value.args[0]


@pytest.mark.skipif(not HAS_H5PY, reason="requires h5py")
def test_read_notable_nopath(tmp_path):
    test_file = tmp_path / "test.hdf5"
    h5py.File(test_file, "w").close()  # create empty file
    with pytest.raises(ValueError, match="no table found in HDF5 group /"):
        Table.read(test_file, path="/", format="hdf5")


@pytest.mark.skipif(not HAS_H5PY, reason="requires h5py")
def test_read_nopath(tmp_path):
    test_file = tmp_path / "test.hdf5"
    t1 = Table()
    t1.add_column(Column(name="a", data=[1, 2, 3]))
    t1.write(test_file, path="the_table")
    t2 = Table.read(test_file)

    assert np.all(t1["a"] == t2["a"])


@pytest.mark.skipif(not HAS_H5PY, reason="requires h5py")
def test_read_nopath_multi_tables(tmp_path):
    test_file = tmp_path / "test.hdf5"
    t1 = Table()
    t1.add_column(Column(name="a", data=[1, 2, 3]))
    t1.write(test_file, path="the_table")
    t1.write(test_file, path="the_table_but_different", append=True, overwrite=True)
    with pytest.warns(
        AstropyUserWarning, match=r"path= was not specified but multiple tables"
    ):
        t2 = Table.read(test_file)

    assert np.all(t1["a"] == t2["a"])


@pytest.mark.skipif(not HAS_H5PY, reason="requires h5py")
def test_write_invalid_path(tmp_path):
    test_file = tmp_path / "test.hdf5"
    t1 = Table()
    t1.add_column(Column(name="a", data=[1, 2, 3]))
    with pytest.raises(ValueError) as exc:
        t1.write(test_file, path="test/")
    assert exc.value.args[0] == "table path should end with table name, not /"


@pytest.mark.skipif(not HAS_H5PY, reason="requires h5py")
def test_read_invalid_path(tmp_path):
    test_file = tmp_path / "test.hdf5"
    t1 = Table()
    t1.add_column(Column(name="a", data=[1, 2, 3]))
    t1.write(test_file, path="the_table")
    with pytest.raises(OSError) as exc:
        Table.read(test_file, path="test/")
    assert exc.value.args[0] == "Path test/ does not exist"


@pytest.mark.skipif(not HAS_H5PY, reason="requires h5py")
def test_read_missing_group(tmp_path):
    test_file = tmp_path / "test.hdf5"
    h5py.File(test_file, "w").close()  # create empty file
    with pytest.raises(OSError) as exc:
        Table.read(test_file, path="test/path/table")
    assert exc.value.args[0] == "Path test/path/table does not exist"


@pytest.mark.skipif(not HAS_H5PY, reason="requires h5py")
def test_read_missing_table(tmp_path):
    test_file = tmp_path / "test.hdf5"
    with h5py.File(test_file, "w") as f:
        f.create_group("test").create_group("path")
    with pytest.raises(OSError) as exc:
        Table.read(test_file, path="test/path/table")
    assert exc.value.args[0] == "Path test/path/table does not exist"


@pytest.mark.skipif(not HAS_H5PY, reason="requires h5py")
def test_read_missing_group_fileobj(tmp_path):
    test_file = tmp_path / "test.hdf5"
    with h5py.File(test_file, "w") as f:
        with pytest.raises(OSError) as exc:
            Table.read(f, path="test/path/table")
        assert exc.value.args[0] == "Path test/path/table does not exist"


@pytest.mark.skipif(not HAS_H5PY, reason="requires h5py")
def test_read_write_simple(tmp_path):
    test_file = tmp_path / "test.hdf5"
    t1 = Table()
    t1.add_column(Column(name="a", data=[1, 2, 3]))
    t1.write(test_file, path="the_table")
    t2 = Table.read(test_file, path="the_table")
    assert np.all(t2["a"] == [1, 2, 3])


@pytest.mark.skipif(not HAS_H5PY, reason="requires h5py")
def test_read_write_existing_table(tmp_path):
    test_file = tmp_path / "test.hdf5"
    t1 = Table()
    t1.add_column(Column(name="a", data=[1, 2, 3]))
    t1.write(test_file, path="the_table")
    with pytest.raises(OSError) as exc:
        t1.write(test_file, path="the_table", append=True)
    assert exc.value.args[0] == "Table the_table already exists"


@pytest.mark.skipif(not HAS_H5PY, reason="requires h5py")
def test_read_write_memory():
    with h5py.File("test", "w", driver="core", backing_store=False) as output_file:
        t1 = Table()
        t1.add_column(Column(name="a", data=[1, 2, 3]))
        t1.write(output_file, path="the_table")
        t2 = Table.read(output_file, path="the_table")
        assert np.all(t2["a"] == [1, 2, 3])


@pytest.mark.skipif(not HAS_H5PY, reason="requires h5py")
def test_read_write_existing(tmp_path):
    test_file = tmp_path / "test.hdf5"
    h5py.File(test_file, "w").close()  # create empty file
    t1 = Table()
    t1.add_column(Column(name="a", data=[1, 2, 3]))
    with pytest.raises(OSError, match=_NOT_OVERWRITING_MSG_MATCH):
        t1.write(test_file, path="the_table")


@pytest.mark.skipif(not HAS_H5PY, reason="requires h5py")
def test_read_write_existing_overwrite(tmp_path):
    test_file = tmp_path / "test.hdf5"
    h5py.File(test_file, "w").close()  # create empty file
    t1 = Table()
    t1.add_column(Column(name="a", data=[1, 2, 3]))
    t1.write(test_file, path="the_table", overwrite=True)
    t2 = Table.read(test_file, path="the_table")
    assert np.all(t2["a"] == [1, 2, 3])


@pytest.mark.skipif(not HAS_H5PY, reason="requires h5py")
def test_read_write_existing_append(tmp_path):
    test_file = tmp_path / "test.hdf5"
    h5py.File(test_file, "w").close()  # create empty file
    t1 = Table()
    t1.add_column(Column(name="a", data=[1, 2, 3]))
    t1.write(test_file, path="the_table_1", append=True)
    t1.write(test_file, path="the_table_2", append=True)
    t2 = Table.read(test_file, path="the_table_1")
    assert np.all(t2["a"] == [1, 2, 3])
    t3 = Table.read(test_file, path="the_table_2")
    assert np.all(t3["a"] == [1, 2, 3])


@pytest.mark.skipif(not HAS_H5PY, reason="requires h5py")
def test_read_write_existing_append_groups(tmp_path):
    test_file = tmp_path / "test.hdf5"
    with h5py.File(test_file, "w") as f:
        f.create_group("test_1")
    t1 = Table()
    t1.add_column(Column(name="a", data=[1, 2, 3]))
    t1.write(test_file, path="test_1/the_table_1", append=True)
    t1.write(test_file, path="test_2/the_table_2", append=True)
    t2 = Table.read(test_file, path="test_1/the_table_1")
    assert np.all(t2["a"] == [1, 2, 3])
    t3 = Table.read(test_file, path="test_2/the_table_2")
    assert np.all(t3["a"] == [1, 2, 3])


@pytest.mark.skipif(not HAS_H5PY, reason="requires h5py")
def test_read_write_existing_append_overwrite(tmp_path):
    test_file = tmp_path / "test.hdf5"
    t1 = Table()
    t1.add_column(Column(name="a", data=[1, 2, 3]))
    t1.write(test_file, path="table1")
    t1.write(test_file, path="table2", append=True)
    t1v2 = Table()
    t1v2.add_column(Column(name="a", data=[4, 5, 6]))
    with pytest.raises(OSError) as exc:
        t1v2.write(test_file, path="table1", append=True)
    assert exc.value.args[0] == "Table table1 already exists"
    t1v2.write(test_file, path="table1", append=True, overwrite=True)
    t2 = Table.read(test_file, path="table1")
    assert np.all(t2["a"] == [4, 5, 6])
    t3 = Table.read(test_file, path="table2")
    assert np.all(t3["a"] == [1, 2, 3])


@pytest.mark.skipif(not HAS_H5PY, reason="requires h5py")
def test_read_fileobj(tmp_path):
    test_file = tmp_path / "test.hdf5"

    t1 = Table()
    t1.add_column(Column(name="a", data=[1, 2, 3]))
    t1.write(test_file, path="the_table")

    import h5py

    with h5py.File(test_file, "r") as input_file:
        t2 = Table.read(input_file, path="the_table")
        assert np.all(t2["a"] == [1, 2, 3])


@pytest.mark.skipif(not HAS_H5PY, reason="requires h5py")
def test_read_filobj_path(tmp_path):
    test_file = tmp_path / "test.hdf5"

    t1 = Table()
    t1.add_column(Column(name="a", data=[1, 2, 3]))
    t1.write(test_file, path="path/to/data/the_table")

    import h5py

    with h5py.File(test_file, "r") as input_file:
        t2 = Table.read(input_file, path="path/to/data/the_table")
        assert np.all(t2["a"] == [1, 2, 3])


@pytest.mark.skipif(not HAS_H5PY, reason="requires h5py")
def test_read_filobj_group_path(tmp_path):
    test_file = tmp_path / "test.hdf5"

    t1 = Table()
    t1.add_column(Column(name="a", data=[1, 2, 3]))
    t1.write(test_file, path="path/to/data/the_table")

    import h5py

    with h5py.File(test_file, "r") as input_file:
        t2 = Table.read(input_file["path/to"], path="data/the_table")
        assert np.all(t2["a"] == [1, 2, 3])


@pytest.mark.skipif(not HAS_H5PY, reason="requires h5py")
def test_read_wrong_fileobj():
    class FakeFile:
        def read(self):
            pass

    f = FakeFile()

    with pytest.raises(TypeError, match="h5py can only open regular files"):
        Table.read(f, format="hdf5")


@pytest.mark.skipif(not HAS_H5PY, reason="requires h5py")
def test_write_fileobj(tmp_path):
    test_file = tmp_path / "test.hdf5"

    import h5py

    with h5py.File(test_file, "w") as output_file:
        t1 = Table()
        t1.add_column(Column(name="a", data=[1, 2, 3]))
        t1.write(output_file, path="the_table")

    t2 = Table.read(test_file, path="the_table")
    assert np.all(t2["a"] == [1, 2, 3])


@pytest.mark.skipif(not HAS_H5PY, reason="requires h5py")
def test_write_create_dataset_kwargs(tmp_path):
    test_file = tmp_path / "test.hdf5"
    the_path = "the_table"

    import h5py

    with h5py.File(test_file, "w") as output_file:
        t1 = Table()
        t1.add_column(Column(name="a", data=[1, 2, 3]))
        t1.write(output_file, path=the_path, maxshape=(None,))

    # A roundabout way of checking this, but the table created above should be
    # resizable if the kwarg was passed through successfully
    t2 = Table()
    t2.add_column(Column(name="a", data=[4, 5]))
    with h5py.File(test_file, "a") as output_file:
        output_file[the_path].resize((len(t1) + len(t2),))
        output_file[the_path][len(t1) :] = t2.as_array()

    t3 = Table.read(test_file, path="the_table")
    assert np.all(t3["a"] == [1, 2, 3, 4, 5])


@pytest.mark.skipif(not HAS_H5PY, reason="requires h5py")
def test_write_filobj_group(tmp_path):
    test_file = tmp_path / "test.hdf5"

    import h5py

    with h5py.File(test_file, "w") as output_file:
        t1 = Table()
        t1.add_column(Column(name="a", data=[1, 2, 3]))
        t1.write(output_file, path="path/to/data/the_table")

    t2 = Table.read(test_file, path="path/to/data/the_table")
    assert np.all(t2["a"] == [1, 2, 3])


@pytest.mark.skipif(not HAS_H5PY, reason="requires h5py")
def test_write_wrong_type():
    t1 = Table()
    t1.add_column(Column(name="a", data=[1, 2, 3]))
    with pytest.raises(TypeError) as exc:
        t1.write(1212, path="path/to/data/the_table", format="hdf5")
    assert (
        exc.value.args[0] == "output should be a string or an h5py File or Group object"
    )


@pytest.mark.skipif(not HAS_H5PY, reason="requires h5py")
@pytest.mark.parametrize("dtype", ALL_DTYPES)
def test_preserve_single_dtypes(tmp_path, dtype):
    test_file = tmp_path / "test.hdf5"

    values = _default_values(dtype)

    t1 = Table()
    t1.add_column(Column(name="a", data=np.array(values, dtype=dtype)))
    t1.write(test_file, path="the_table")

    t2 = Table.read(test_file, path="the_table")

    assert np.all(t2["a"] == values)
    assert t2["a"].dtype == dtype


@pytest.mark.skipif(not HAS_H5PY, reason="requires h5py")
def test_preserve_all_dtypes(tmp_path):
    test_file = tmp_path / "test.hdf5"

    t1 = Table()

    for dtype in ALL_DTYPES:
        values = _default_values(dtype)
        t1.add_column(Column(name=str(dtype), data=np.array(values, dtype=dtype)))

    t1.write(test_file, path="the_table")

    t2 = Table.read(test_file, path="the_table")

    for dtype in ALL_DTYPES:
        values = _default_values(dtype)
        assert np.all(t2[str(dtype)] == values)
        assert t2[str(dtype)].dtype == dtype


@pytest.mark.skipif(not HAS_H5PY, reason="requires h5py")
def test_preserve_meta(tmp_path):
    test_file = tmp_path / "test.hdf5"

    t1 = Table()
    t1.add_column(Column(name="a", data=[1, 2, 3]))

    t1.meta["a"] = 1
    t1.meta["b"] = "hello"
    t1.meta["c"] = 3.14159
    t1.meta["d"] = True
    t1.meta["e"] = np.array([1, 2, 3])

    t1.write(test_file, path="the_table")

    t2 = Table.read(test_file, path="the_table")

    for key in t1.meta:
        assert np.all(t1.meta[key] == t2.meta[key])


@pytest.mark.skipif(not HAS_H5PY, reason="requires h5py")
def test_preserve_serialized(tmp_path):
    test_file = tmp_path / "test.hdf5"

    t1 = Table()
    t1["a"] = Column(data=[1, 2, 3], unit="s")
    t1["a"].meta["a0"] = "A0"
    t1["a"].meta["a1"] = {"a1": [0, 1]}
    t1["a"].format = "7.3f"
    t1["a"].description = "A column"
    t1.meta["b"] = 1
    t1.meta["c"] = {"c0": [0, 1]}

    t1.write(test_file, path="the_table", serialize_meta=True, overwrite=True)

    t2 = Table.read(test_file, path="the_table")

    assert t1["a"].unit == t2["a"].unit
    assert t1["a"].format == t2["a"].format
    assert t1["a"].description == t2["a"].description
    assert t1["a"].meta == t2["a"].meta
    assert t1.meta == t2.meta

    # Check that the meta table is fixed-width bytes (see #11299)
    h5 = h5py.File(test_file, "r")
    meta_lines = h5[meta_path("the_table")]
    assert meta_lines.dtype.kind == "S"


@pytest.mark.skipif(not HAS_H5PY, reason="requires h5py")
def test_preserve_serialized_old_meta_format():
    """Test the old meta format

    Only for some files created prior to v4.0, in compatibility mode.
    """
    test_file = get_pkg_data_filename("data/old_meta_example.hdf5")

    t1 = Table()
    t1["a"] = Column(data=[1, 2, 3], unit="s")
    t1["a"].meta["a0"] = "A0"
    t1["a"].meta["a1"] = {"a1": [0, 1]}
    t1["a"].format = "7.3f"
    t1["a"].description = "A column"
    t1.meta["b"] = 1
    t1.meta["c"] = {"c0": [0, 1]}

    t2 = Table.read(test_file, path="the_table")

    assert t1["a"].unit == t2["a"].unit
    assert t1["a"].format == t2["a"].format
    assert t1["a"].description == t2["a"].description
    assert t1["a"].meta == t2["a"].meta
    assert t1.meta == t2.meta


@pytest.mark.skipif(not HAS_H5PY, reason="requires h5py")
def test_preserve_serialized_in_complicated_path(tmp_path):
    test_file = tmp_path / "test.hdf5"

    t1 = Table()
    t1["a"] = Column(data=[1, 2, 3], unit="s")
    t1["a"].meta["a0"] = "A0"
    t1["a"].meta["a1"] = {"a1": [0, 1]}
    t1["a"].format = "7.3f"
    t1["a"].description = "A column"
    t1.meta["b"] = 1
    t1.meta["c"] = {"c0": [0, 1]}

    t1.write(
        test_file,
        path="the_table/complicated/path",
        serialize_meta=True,
        overwrite=True,
    )

    t2 = Table.read(test_file, path="the_table/complicated/path")

    assert t1["a"].format == t2["a"].format
    assert t1["a"].unit == t2["a"].unit
    assert t1["a"].description == t2["a"].description
    assert t1["a"].meta == t2["a"].meta
    assert t1.meta == t2.meta


@pytest.mark.skipif(not HAS_H5PY, reason="requires h5py")
def test_metadata_very_large(tmp_path):
    """Test that very large datasets work, now!"""
    test_file = tmp_path / "test.hdf5"

    t1 = Table()
    t1["a"] = Column(data=[1, 2, 3], unit="s")
    t1["a"].meta["a0"] = "A0"
    t1["a"].meta["a1"] = {"a1": [0, 1]}
    t1["a"].format = "7.3f"
    t1["a"].description = "A column"
    t1.meta["b"] = 1
    t1.meta["c"] = {"c0": [0, 1]}
    t1.meta["meta_big"] = "0" * (2**16 + 1)
    t1.meta["meta_biggerstill"] = "0" * (2**18)

    t1.write(test_file, path="the_table", serialize_meta=True, overwrite=True)

    t2 = Table.read(test_file, path="the_table")

    assert t1["a"].unit == t2["a"].unit
    assert t1["a"].format == t2["a"].format
    assert t1["a"].description == t2["a"].description
    assert t1["a"].meta == t2["a"].meta
    assert t1.meta == t2.meta


@pytest.mark.skipif(not HAS_H5PY, reason="requires h5py")
def test_skip_meta(tmp_path):
    test_file = tmp_path / "test.hdf5"

    t1 = Table()
    t1.add_column(Column(name="a", data=[1, 2, 3]))

    t1.meta["a"] = 1
    t1.meta["b"] = "hello"
    t1.meta["c"] = 3.14159
    t1.meta["d"] = True
    t1.meta["e"] = np.array([1, 2, 3])
    t1.meta["f"] = str

    wtext = (
        f"Attribute `f` of type {type(t1.meta['f'])} cannot be written to HDF5 files -"
        " skipping"
    )
    with pytest.warns(AstropyUserWarning, match=wtext) as w:
        t1.write(test_file, path="the_table")
    assert len(w) == 1


@pytest.mark.skipif(not HAS_H5PY, reason="requires h5py")
def test_fail_meta_serialize(tmp_path):
    test_file = tmp_path / "test.hdf5"

    t1 = Table()
    t1.add_column(Column(name="a", data=[1, 2, 3]))
    t1.meta["f"] = str

    with pytest.raises(Exception) as err:
        t1.write(test_file, path="the_table", serialize_meta=True)
    assert "cannot represent an object" in str(err.value)
    assert "<class 'str'>" in str(err.value)


@pytest.mark.skipif(not HAS_H5PY, reason="requires h5py")
def test_read_h5py_objects(tmp_path):
    # Regression test - ensure that Datasets are recognized automatically

    test_file = tmp_path / "test.hdf5"

    import h5py

    with h5py.File(test_file, "w") as output_file:
        t1 = Table()
        t1.add_column(Column(name="a", data=[1, 2, 3]))
        t1.write(output_file, path="the_table")

    f = h5py.File(test_file, mode="r")

    t2 = Table.read(f, path="the_table")
    assert np.all(t2["a"] == [1, 2, 3])

    t3 = Table.read(f["/"], path="the_table")
    assert np.all(t3["a"] == [1, 2, 3])

    t4 = Table.read(f["the_table"])
    assert np.all(t4["a"] == [1, 2, 3])

    f.close()  # don't leave the file open


@pytest.mark.skipif(not HAS_H5PY, reason="requires h5py")
def test_read_write_unicode_to_hdf5(tmp_path):
    test_file = tmp_path / "test.hdf5"

    t = Table()
    t["p"] = ["a", "b", "c"]
    t["q"] = [1, 2, 3]
    t["r"] = [b"a", b"b", b"c"]
    t["s"] = ["\u2119", "\u01b4", "\u2602"]
    t.write(test_file, path="the_table", overwrite=True)

    t1 = Table.read(test_file, path="the_table", character_as_bytes=False)
    for col, col1 in zip(t.itercols(), t1.itercols()):
        assert np.all(col == col1)
    assert np.all(t1["p"].info.dtype.kind == "U")
    assert np.all(t1["q"].info.dtype.kind == "i")
    assert np.all(t1["r"].info.dtype.kind == "U")
    assert np.all(t1["s"].info.dtype.kind == "U")

    # Test default (character_as_bytes=True)
    t2 = Table.read(test_file, path="the_table")
    for col, col1 in zip(t.itercols(), t2.itercols()):
        assert np.all(col == col1)
    assert np.all(t2["p"].info.dtype.kind == "S")
    assert np.all(t2["q"].info.dtype.kind == "i")
    assert np.all(t2["r"].info.dtype.kind == "S")
    assert np.all(t2["s"].info.dtype.kind == "S")


def assert_objects_equal(obj1, obj2, attrs, compare_class=True):
    if compare_class:
        assert obj1.__class__ is obj2.__class__

    info_attrs = [
        "info.name",
        "info.format",
        "info.unit",
        "info.description",
        "info.meta",
        "info.dtype",
    ]
    for attr in attrs + info_attrs:
        a1 = obj1
        a2 = obj2
        for subattr in attr.split("."):
            try:
                a1 = getattr(a1, subattr)
                a2 = getattr(a2, subattr)
            except AttributeError:
                a1 = a1[subattr]
                a2 = a2[subattr]

        # Mixin info.meta can None instead of empty OrderedDict(), #6720 would
        # fix this.
        if attr == "info.meta":
            if a1 is None:
                a1 = {}
            if a2 is None:
                a2 = {}

        if isinstance(a1, np.ndarray) and a1.dtype.kind == "f":
            assert quantity_allclose(a1, a2, rtol=1e-15)
        elif isinstance(a1, np.dtype):
            # HDF5 does not perfectly preserve dtype: byte order can change, and
            # unicode gets stored as bytes.  So, we just check safe casting, to
            # ensure we do not, e.g., accidentally change integer to float, etc.
            if NUMPY_LT_1_22 and a1.names:
                # For old numpy, can_cast does not deal well with structured dtype.
                assert a1.names == a2.names
            else:
                assert np.can_cast(a2, a1, casting="safe")
        else:
            assert np.all(a1 == a2)


@pytest.mark.skipif(not HAS_H5PY, reason="requires h5py")
def test_hdf5_mixins_qtable_to_table(tmp_path):
    """Test writing as QTable and reading as Table.  Ensure correct classes
    come out.
    """
    filename = tmp_path / "test_simple.hdf5"

    names = sorted(mixin_cols)

    t = QTable([mixin_cols[name] for name in names], names=names)
    t.write(filename, format="hdf5", path="root", serialize_meta=True)
    t2 = Table.read(filename, format="hdf5", path="root")

    assert t.colnames == t2.colnames

    for name, col in t.columns.items():
        col2 = t2[name]

        attrs = compare_attrs[name]
        compare_class = True

        if isinstance(col.info, QuantityInfo):
            # Downgrade Quantity to Column + unit
            assert type(col2) is Column
            # Class-specific attributes like `value` or `wrap_angle` are lost.
            attrs = ["unit"]
            compare_class = False
            # Compare data values here (assert_objects_equal doesn't know how in this case)
            assert np.all(col.value == col2)

        assert_objects_equal(col, col2, attrs, compare_class)


@pytest.mark.skipif(not HAS_H5PY, reason="requires h5py")
@pytest.mark.parametrize("table_cls", (Table, QTable))
def test_hdf5_mixins_as_one(table_cls, tmp_path):
    """Test write/read all cols at once and validate intermediate column names"""
    filename = tmp_path / "test_simple.hdf5"
    names = sorted(mixin_cols)
    all_serialized_names = []
    for name in names:
        all_serialized_names.extend(serialized_names[name])

    t = table_cls([mixin_cols[name] for name in names], names=names)
    t.meta["C"] = "spam"
    t.meta["comments"] = ["this", "is", "a", "comment"]
    t.meta["history"] = ["first", "second", "third"]

    t.write(filename, format="hdf5", path="root", serialize_meta=True)

    t2 = table_cls.read(filename, format="hdf5", path="root")
    assert t2.meta["C"] == "spam"
    assert t2.meta["comments"] == ["this", "is", "a", "comment"]
    assert t2.meta["history"] == ["first", "second", "third"]

    assert t.colnames == t2.colnames

    # Read directly via hdf5 and confirm column names
    h5 = h5py.File(filename, "r")
    h5_names = list(h5["root"].dtype.names)
    assert h5_names == all_serialized_names
    h5.close()


@pytest.mark.skipif(not HAS_H5PY, reason="requires h5py")
@pytest.mark.parametrize("name_col", list(mixin_cols.items()))
@pytest.mark.parametrize("table_cls", (Table, QTable))
def test_hdf5_mixins_per_column(table_cls, name_col, tmp_path):
    """Test write/read one col at a time and do detailed validation"""
    filename = tmp_path / "test_simple.hdf5"
    name, col = name_col

    c = [1.0, 2.0]
    t = table_cls([c, col, c], names=["c1", name, "c2"])
    t[name].info.description = "my description"
    t[name].info.meta = {"list": list(range(50)), "dict": {"a": "b" * 200}}

    if not t.has_mixin_columns:
        pytest.skip("column is not a mixin (e.g. Quantity subclass in Table)")

    t.write(filename, format="hdf5", path="root", serialize_meta=True)
    t2 = table_cls.read(filename, format="hdf5", path="root")

    assert t.colnames == t2.colnames

    for colname in t.colnames:
        compare = ["data"] if colname in ("c1", "c2") else compare_attrs[colname]
        assert_objects_equal(t[colname], t2[colname], compare)

    # Special case to make sure Column type doesn't leak into Time class data
    if name.startswith("tm"):
        assert t2[name]._time.jd1.__class__ is np.ndarray
        assert t2[name]._time.jd2.__class__ is np.ndarray


@pytest.mark.parametrize("name_col", unsupported_cols.items())
@pytest.mark.xfail(reason="column type unsupported")
def test_fits_unsupported_mixin(self, name_col, tmp_path):
    # Check that we actually fail in writing unsupported columns defined
    # on top.
    filename = tmp_path / "test_simple.fits"
    name, col = name_col
    Table([col], names=[name]).write(
        filename, format="hdf5", path="root", serialize_meta=True
    )


@pytest.mark.skipif(not HAS_H5PY, reason="requires h5py")
def test_round_trip_masked_table_default(tmp_path):
    """Test round-trip of MaskedColumn through HDF5 using default serialization
    that writes a separate mask column.  Note:

    >>> simple_table(masked=True)
    <Table masked=True length=3>
      a      b     c
    int64 float64 str1
    ----- ------- ----
       --     1.0    c
        2     2.0   --
        3      --    e
    """
    filename = tmp_path / "test.h5"

    t = simple_table(masked=True)  # int, float, and str cols with one masked element
    t["c"] = [b"c", b"d", b"e"]
    t["c"].mask[1] = True
    t.write(filename, format="hdf5", path="root", serialize_meta=True)

    t2 = Table.read(filename)
    assert t2.masked is False
    assert t2.colnames == t.colnames
    for name in t2.colnames:
        assert np.all(t2[name].mask == t[name].mask)
        assert np.all(t2[name] == t[name])

        # Data under the mask round-trips also (unmask data to show this).
        t[name].mask = False
        t2[name].mask = False
        assert np.all(t2[name] == t[name])


@pytest.mark.skipif(not HAS_H5PY, reason="requires h5py")
def test_overwrite_serialized_meta():
    # This used to cause an error because the meta data table
    # was not removed from the existing file.

    with h5py.File("test_data.h5", "w", driver="core", backing_store=False) as out:
        t1 = Table()
        t1.add_column(Column(data=[4, 8, 15], unit="cm"))
        t1.write(out, path="data", serialize_meta=True)

        t2 = Table.read(out, path="data")
        assert all(t1 == t2)
        assert t1.info(out=None) == t2.info(out=None)

        t3 = Table()
        t3.add_column(Column(data=[16, 23, 42], unit="g"))
        t3.write(out, path="data", serialize_meta=True, append=True, overwrite=True)

        t2 = Table.read(out, path="data")
        assert all(t3 == t2)
        assert t3.info(out=None) == t2.info(out=None)


@pytest.mark.skipif(not HAS_H5PY, reason="requires h5py")
def test_read_write_tilde_path(home_is_tmpdir):
    test_file = os.path.join("~", "test.hdf5")
    t1 = Table()
    t1["a"] = [1, 2, 3]

    t1.write(test_file, path="the_table")

    t1 = Table.read(test_file, path="the_table")
    t1 = Table.read(test_file, path="the_table", format="hdf5")

    # Ensure the data wasn't written to the literal tilde-prefixed path
    assert not os.path.exists(test_file)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""Checks for optional dependencies using lazy import from
`PEP 562 <https://www.python.org/dev/peps/pep-0562/>`_.
"""
import importlib
import warnings

# First, the top-level packages:
# TODO: This list is a duplicate of the dependencies in setup.cfg "all", but
# some of the package names are different from the pip-install name (e.g.,
# beautifulsoup4 -> bs4).
_optional_deps = [
    "asdf",
    "asdf_astropy",
    "bleach",
    "bottleneck",
    "bs4",
    "bz2",
    "fsspec",
    "h5py",
    "html5lib",
    "IPython",
    "jplephem",
    "lxml",
    "matplotlib",
    "mpmath",
    "pandas",
    "PIL",
    "pytz",
    "s3fs",
    "scipy",
    "skyfield",
    "sortedcontainers",
    "lzma",
    "pyarrow",
    "pytest_mpl",
]
_formerly_optional_deps = ["yaml"]  # for backward compatibility
_deps = {k.upper(): k for k in _optional_deps + _formerly_optional_deps}

# Any subpackages that have different import behavior:
_deps["PLT"] = "matplotlib.pyplot"

__all__ = [f"HAS_{pkg}" for pkg in _deps]


def __getattr__(name):
    if name in __all__:
        module_name = name[4:]

        if module_name == "YAML":
            from astropy.utils.exceptions import AstropyDeprecationWarning

            warnings.warn(
                "PyYaml is now a strict dependency. HAS_YAML is deprecated as "
                "of v5.0 and will be removed in a subsequent version.",
                category=AstropyDeprecationWarning,
            )

        try:
            importlib.import_module(_deps[module_name])
        except (ImportError, ModuleNotFoundError):
            return False
        return True

    raise AttributeError(f"Module {__name__!r} has no attribute {name!r}.")



from astropy.io.misc.asdf.tags.transform.basic import TransformType
from astropy.modeling.models import Spline1D

__all__ = ["SplineType"]


class SplineType(TransformType):
    name = "transform/spline1d"
    version = "1.0.0"
    types = ["astropy.modeling.spline.Spline1D"]

    @classmethod
    def from_tree_transform(cls, node, ctx):
        return Spline1D(
            knots=node["knots"], coeffs=node["coefficients"], degree=node["degree"]
        )

    @classmethod
    def to_tree_transform(cls, model, ctx):
        return {"knots": model.t, "coefficients": model.c, "degree": model.degree}



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""Accuracy tests for GCRS coordinate transformations, primarily to/from AltAz.

"""
import os
import warnings
from importlib import metadata

import erfa
import numpy as np
import pytest

from astropy import units as u
from astropy.coordinates import (
    CIRS,
    GCRS,
    HCRS,
    ICRS,
    ITRS,
    TEME,
    TETE,
    AltAz,
    CartesianDifferential,
    CartesianRepresentation,
    EarthLocation,
    HADec,
    HeliocentricMeanEcliptic,
    PrecessedGeocentric,
    SkyCoord,
    SphericalRepresentation,
    UnitSphericalRepresentation,
    get_sun,
    solar_system_ephemeris,
)
from astropy.coordinates.angle_utilities import golden_spiral_grid
from astropy.coordinates.builtin_frames.intermediate_rotation_transforms import (
    cirs_to_itrs_mat,
    gcrs_to_cirs_mat,
    get_location_gcrs,
    tete_to_itrs_mat,
)
from astropy.coordinates.builtin_frames.utils import get_jd12
from astropy.coordinates.solar_system import (
    _apparent_position_in_true_coordinates,
    get_body,
)
from astropy.tests.helper import assert_quantity_allclose as assert_allclose
from astropy.time import Time
from astropy.units import allclose
from astropy.utils import iers
from astropy.utils.compat.optional_deps import HAS_JPLEPHEM
from astropy.utils.exceptions import AstropyDeprecationWarning, AstropyWarning

CI = os.environ.get("CI", False) == "true"


def test_icrs_cirs():
    """
    Check a few cases of ICRS<->CIRS for consistency.

    Also includes the CIRS<->CIRS transforms at different times, as those go
    through ICRS
    """
    usph = golden_spiral_grid(200)
    dist = np.linspace(0.0, 1, len(usph)) * u.pc
    inod = ICRS(usph)
    iwd = ICRS(ra=usph.lon, dec=usph.lat, distance=dist)

    cframe1 = CIRS()
    cirsnod = inod.transform_to(cframe1)  # uses the default time
    # first do a round-tripping test
    inod2 = cirsnod.transform_to(ICRS())
    assert_allclose(inod.ra, inod2.ra)
    assert_allclose(inod.dec, inod2.dec)

    # now check that a different time yields different answers
    cframe2 = CIRS(obstime=Time("J2005"))
    cirsnod2 = inod.transform_to(cframe2)
    assert not allclose(cirsnod.ra, cirsnod2.ra, rtol=1e-8)
    assert not allclose(cirsnod.dec, cirsnod2.dec, rtol=1e-8)

    # parallax effects should be included, so with and w/o distance should be different
    cirswd = iwd.transform_to(cframe1)
    assert not allclose(cirswd.ra, cirsnod.ra, rtol=1e-8)
    assert not allclose(cirswd.dec, cirsnod.dec, rtol=1e-8)
    # and the distance should transform at least somehow
    assert not allclose(cirswd.distance, iwd.distance, rtol=1e-8)

    # now check that the cirs self-transform works as expected
    cirsnod3 = cirsnod.transform_to(cframe1)  # should be a no-op
    assert_allclose(cirsnod.ra, cirsnod3.ra)
    assert_allclose(cirsnod.dec, cirsnod3.dec)

    cirsnod4 = cirsnod.transform_to(cframe2)  # should be different
    assert not allclose(cirsnod4.ra, cirsnod.ra, rtol=1e-8)
    assert not allclose(cirsnod4.dec, cirsnod.dec, rtol=1e-8)

    cirsnod5 = cirsnod4.transform_to(cframe1)  # should be back to the same
    assert_allclose(cirsnod.ra, cirsnod5.ra)
    assert_allclose(cirsnod.dec, cirsnod5.dec)


usph = golden_spiral_grid(200)
dist = np.linspace(0.5, 1, len(usph)) * u.pc
icrs_coords = [ICRS(usph), ICRS(usph.lon, usph.lat, distance=dist)]
gcrs_frames = [GCRS(), GCRS(obstime=Time("J2005"))]


@pytest.mark.parametrize("icoo", icrs_coords)
def test_icrs_gcrs(icoo):
    """
    Check ICRS<->GCRS for consistency
    """
    gcrscoo = icoo.transform_to(gcrs_frames[0])  # uses the default time
    # first do a round-tripping test
    icoo2 = gcrscoo.transform_to(ICRS())
    assert_allclose(icoo.distance, icoo2.distance)
    assert_allclose(icoo.ra, icoo2.ra)
    assert_allclose(icoo.dec, icoo2.dec)
    assert isinstance(icoo2.data, icoo.data.__class__)

    # now check that a different time yields different answers
    gcrscoo2 = icoo.transform_to(gcrs_frames[1])
    assert not allclose(gcrscoo.ra, gcrscoo2.ra, rtol=1e-8, atol=1e-10 * u.deg)
    assert not allclose(gcrscoo.dec, gcrscoo2.dec, rtol=1e-8, atol=1e-10 * u.deg)

    # now check that the cirs self-transform works as expected
    gcrscoo3 = gcrscoo.transform_to(gcrs_frames[0])  # should be a no-op
    assert_allclose(gcrscoo.ra, gcrscoo3.ra)
    assert_allclose(gcrscoo.dec, gcrscoo3.dec)

    gcrscoo4 = gcrscoo.transform_to(gcrs_frames[1])  # should be different
    assert not allclose(gcrscoo4.ra, gcrscoo.ra, rtol=1e-8, atol=1e-10 * u.deg)
    assert not allclose(gcrscoo4.dec, gcrscoo.dec, rtol=1e-8, atol=1e-10 * u.deg)

    gcrscoo5 = gcrscoo4.transform_to(gcrs_frames[0])  # should be back to the same
    assert_allclose(gcrscoo.ra, gcrscoo5.ra, rtol=1e-8, atol=1e-10 * u.deg)
    assert_allclose(gcrscoo.dec, gcrscoo5.dec, rtol=1e-8, atol=1e-10 * u.deg)

    # also make sure that a GCRS with a different geoloc/geovel gets a different answer
    # roughly a moon-like frame
    gframe3 = GCRS(obsgeoloc=[385000.0, 0, 0] * u.km, obsgeovel=[1, 0, 0] * u.km / u.s)
    gcrscoo6 = icoo.transform_to(gframe3)  # should be different
    assert not allclose(gcrscoo.ra, gcrscoo6.ra, rtol=1e-8, atol=1e-10 * u.deg)
    assert not allclose(gcrscoo.dec, gcrscoo6.dec, rtol=1e-8, atol=1e-10 * u.deg)
    icooviag3 = gcrscoo6.transform_to(ICRS())  # and now back to the original
    assert_allclose(icoo.ra, icooviag3.ra)
    assert_allclose(icoo.dec, icooviag3.dec)


@pytest.mark.parametrize("gframe", gcrs_frames)
def test_icrs_gcrs_dist_diff(gframe):
    """
    Check that with and without distance give different ICRS<->GCRS answers
    """
    gcrsnod = icrs_coords[0].transform_to(gframe)
    gcrswd = icrs_coords[1].transform_to(gframe)

    # parallax effects should be included, so with and w/o distance should be different
    assert not allclose(gcrswd.ra, gcrsnod.ra, rtol=1e-8, atol=1e-10 * u.deg)
    assert not allclose(gcrswd.dec, gcrsnod.dec, rtol=1e-8, atol=1e-10 * u.deg)
    # and the distance should transform at least somehow
    assert not allclose(
        gcrswd.distance, icrs_coords[1].distance, rtol=1e-8, atol=1e-10 * u.pc
    )


def test_cirs_to_altaz():
    """
    Check the basic CIRS<->AltAz transforms.  More thorough checks implicitly
    happen in `test_iau_fullstack`
    """
    from astropy.coordinates import EarthLocation

    usph = golden_spiral_grid(200)
    dist = np.linspace(0.5, 1, len(usph)) * u.pc
    cirs = CIRS(usph, obstime="J2000")
    crepr = SphericalRepresentation(lon=usph.lon, lat=usph.lat, distance=dist)
    cirscart = CIRS(
        crepr, obstime=cirs.obstime, representation_type=CartesianRepresentation
    )

    loc = EarthLocation(lat=0 * u.deg, lon=0 * u.deg, height=0 * u.m)
    altazframe = AltAz(location=loc, obstime=Time("J2005"))

    cirs2 = cirs.transform_to(altazframe).transform_to(cirs)
    cirs3 = cirscart.transform_to(altazframe).transform_to(cirs)

    # check round-tripping
    assert_allclose(cirs.ra, cirs2.ra)
    assert_allclose(cirs.dec, cirs2.dec)
    assert_allclose(cirs.ra, cirs3.ra)
    assert_allclose(cirs.dec, cirs3.dec)


def test_cirs_to_hadec():
    """
    Check the basic CIRS<->HADec transforms.
    """
    from astropy.coordinates import EarthLocation

    usph = golden_spiral_grid(200)
    dist = np.linspace(0.5, 1, len(usph)) * u.pc
    cirs = CIRS(usph, obstime="J2000")
    crepr = SphericalRepresentation(lon=usph.lon, lat=usph.lat, distance=dist)
    cirscart = CIRS(
        crepr, obstime=cirs.obstime, representation_type=CartesianRepresentation
    )

    loc = EarthLocation(lat=0 * u.deg, lon=0 * u.deg, height=0 * u.m)
    hadecframe = HADec(location=loc, obstime=Time("J2005"))

    cirs2 = cirs.transform_to(hadecframe).transform_to(cirs)
    cirs3 = cirscart.transform_to(hadecframe).transform_to(cirs)

    # check round-tripping
    assert_allclose(cirs.ra, cirs2.ra)
    assert_allclose(cirs.dec, cirs2.dec)
    assert_allclose(cirs.ra, cirs3.ra)
    assert_allclose(cirs.dec, cirs3.dec)


def test_itrs_topo_to_altaz_with_refraction():
    loc = EarthLocation(lat=0 * u.deg, lon=0 * u.deg, height=0 * u.m)
    usph = golden_spiral_grid(200)
    dist = np.linspace(1.0, 1000.0, len(usph)) * u.au
    icrs = ICRS(ra=usph.lon, dec=usph.lat, distance=dist)
    altaz_frame1 = AltAz(obstime="J2000", location=loc)
    altaz_frame2 = AltAz(
        obstime="J2000", location=loc, pressure=1000.0 * u.hPa, relative_humidity=0.5
    )
    cirs_frame = CIRS(obstime="J2000", location=loc)
    itrs_frame = ITRS(location=loc)

    # Normal route
    # No Refraction
    altaz1 = icrs.transform_to(altaz_frame1)

    # Refraction added
    altaz2 = icrs.transform_to(altaz_frame2)

    # Refraction removed
    cirs = altaz2.transform_to(cirs_frame)
    altaz3 = cirs.transform_to(altaz_frame1)

    # Through ITRS
    # No Refraction
    itrs = icrs.transform_to(itrs_frame)
    altaz11 = itrs.transform_to(altaz_frame1)

    assert_allclose(altaz11.az - altaz1.az, 0 * u.mas, atol=0.1 * u.mas)
    assert_allclose(altaz11.alt - altaz1.alt, 0 * u.mas, atol=0.1 * u.mas)
    assert_allclose(altaz11.distance - altaz1.distance, 0 * u.cm, atol=10.0 * u.cm)

    # Round trip
    itrs11 = altaz11.transform_to(itrs_frame)

    assert_allclose(itrs11.x, itrs.x)
    assert_allclose(itrs11.y, itrs.y)
    assert_allclose(itrs11.z, itrs.z)

    # Refraction added
    altaz22 = itrs.transform_to(altaz_frame2)

    assert_allclose(altaz22.az - altaz2.az, 0 * u.mas, atol=0.1 * u.mas)
    assert_allclose(altaz22.alt - altaz2.alt, 0 * u.mas, atol=0.1 * u.mas)
    assert_allclose(altaz22.distance - altaz2.distance, 0 * u.cm, atol=10.0 * u.cm)

    # Refraction removed
    itrs = altaz22.transform_to(itrs_frame)
    altaz33 = itrs.transform_to(altaz_frame1)

    assert_allclose(altaz33.az - altaz3.az, 0 * u.mas, atol=0.1 * u.mas)
    assert_allclose(altaz33.alt - altaz3.alt, 0 * u.mas, atol=0.1 * u.mas)
    assert_allclose(altaz33.distance - altaz3.distance, 0 * u.cm, atol=10.0 * u.cm)


def test_itrs_topo_to_hadec_with_refraction():
    loc = EarthLocation(lat=0 * u.deg, lon=0 * u.deg, height=0 * u.m)
    usph = golden_spiral_grid(200)
    dist = np.linspace(1.0, 1000.0, len(usph)) * u.au
    icrs = ICRS(ra=usph.lon, dec=usph.lat, distance=dist)
    hadec_frame1 = HADec(obstime="J2000", location=loc)
    hadec_frame2 = HADec(
        obstime="J2000", location=loc, pressure=1000.0 * u.hPa, relative_humidity=0.5
    )
    cirs_frame = CIRS(obstime="J2000", location=loc)
    itrs_frame = ITRS(location=loc)

    # Normal route
    # No Refraction
    hadec1 = icrs.transform_to(hadec_frame1)

    # Refraction added
    hadec2 = icrs.transform_to(hadec_frame2)

    # Refraction removed
    cirs = hadec2.transform_to(cirs_frame)
    hadec3 = cirs.transform_to(hadec_frame1)

    # Through ITRS
    # No Refraction
    itrs = icrs.transform_to(itrs_frame)
    hadec11 = itrs.transform_to(hadec_frame1)

    assert_allclose(hadec11.ha - hadec1.ha, 0 * u.mas, atol=0.1 * u.mas)
    assert_allclose(hadec11.dec - hadec1.dec, 0 * u.mas, atol=0.1 * u.mas)
    assert_allclose(hadec11.distance - hadec1.distance, 0 * u.cm, atol=10.0 * u.cm)

    # Round trip
    itrs11 = hadec11.transform_to(itrs_frame)

    assert_allclose(itrs11.x, itrs.x)
    assert_allclose(itrs11.y, itrs.y)
    assert_allclose(itrs11.z, itrs.z)

    # Refraction added
    hadec22 = itrs.transform_to(hadec_frame2)

    assert_allclose(hadec22.ha - hadec2.ha, 0 * u.mas, atol=0.1 * u.mas)
    assert_allclose(hadec22.dec - hadec2.dec, 0 * u.mas, atol=0.1 * u.mas)
    assert_allclose(hadec22.distance - hadec2.distance, 0 * u.cm, atol=10.0 * u.cm)

    # Refraction removed
    itrs = hadec22.transform_to(itrs_frame)
    hadec33 = itrs.transform_to(hadec_frame1)

    assert_allclose(hadec33.ha - hadec3.ha, 0 * u.mas, atol=0.1 * u.mas)
    assert_allclose(hadec33.dec - hadec3.dec, 0 * u.mas, atol=0.1 * u.mas)
    assert_allclose(hadec33.distance - hadec3.distance, 0 * u.cm, atol=10.0 * u.cm)


def test_gcrs_itrs():
    """
    Check basic GCRS<->ITRS transforms for round-tripping.
    """
    usph = golden_spiral_grid(200)
    gcrs = GCRS(usph, obstime="J2000")
    gcrs6 = GCRS(usph, obstime="J2006")

    gcrs2 = gcrs.transform_to(ITRS()).transform_to(gcrs)
    gcrs6_2 = gcrs6.transform_to(ITRS()).transform_to(gcrs)

    assert_allclose(gcrs.ra, gcrs2.ra)
    assert_allclose(gcrs.dec, gcrs2.dec)
    # these should be different:
    assert not allclose(gcrs.ra, gcrs6_2.ra, rtol=1e-8)
    assert not allclose(gcrs.dec, gcrs6_2.dec, rtol=1e-8)

    # also try with the cartesian representation
    gcrsc = gcrs.realize_frame(gcrs.data)
    gcrsc.representation_type = CartesianRepresentation
    gcrsc2 = gcrsc.transform_to(ITRS()).transform_to(gcrsc)
    assert_allclose(gcrsc.spherical.lon, gcrsc2.ra)
    assert_allclose(gcrsc.spherical.lat, gcrsc2.dec)


def test_cirs_itrs():
    """
    Check basic CIRS<->ITRS geocentric transforms for round-tripping.
    """
    usph = golden_spiral_grid(200)
    cirs = CIRS(usph, obstime="J2000")
    cirs6 = CIRS(usph, obstime="J2006")

    cirs2 = cirs.transform_to(ITRS()).transform_to(cirs)
    cirs6_2 = cirs6.transform_to(ITRS()).transform_to(cirs)  # different obstime

    # just check round-tripping
    assert_allclose(cirs.ra, cirs2.ra)
    assert_allclose(cirs.dec, cirs2.dec)
    assert not allclose(cirs.ra, cirs6_2.ra)
    assert not allclose(cirs.dec, cirs6_2.dec)


def test_cirs_itrs_topo():
    """
    Check basic CIRS<->ITRS topocentric transforms for round-tripping.
    """
    loc = EarthLocation(lat=0 * u.deg, lon=0 * u.deg, height=0 * u.m)
    usph = golden_spiral_grid(200)
    cirs = CIRS(usph, obstime="J2000", location=loc)
    cirs6 = CIRS(usph, obstime="J2006", location=loc)

    cirs2 = cirs.transform_to(ITRS(location=loc)).transform_to(cirs)
    # different obstime
    cirs6_2 = cirs6.transform_to(ITRS(location=loc)).transform_to(cirs)

    # just check round-tripping
    assert_allclose(cirs.ra, cirs2.ra)
    assert_allclose(cirs.dec, cirs2.dec)
    assert not allclose(cirs.ra, cirs6_2.ra)
    assert not allclose(cirs.dec, cirs6_2.dec)


def test_gcrs_cirs():
    """
    Check GCRS<->CIRS transforms for round-tripping.  More complicated than the
    above two because it's multi-hop
    """
    usph = golden_spiral_grid(200)
    gcrs = GCRS(usph, obstime="J2000")
    gcrs6 = GCRS(usph, obstime="J2006")

    gcrs2 = gcrs.transform_to(CIRS()).transform_to(gcrs)
    gcrs6_2 = gcrs6.transform_to(CIRS()).transform_to(gcrs)

    assert_allclose(gcrs.ra, gcrs2.ra)
    assert_allclose(gcrs.dec, gcrs2.dec)
    # these should be different:
    assert not allclose(gcrs.ra, gcrs6_2.ra, rtol=1e-8)
    assert not allclose(gcrs.dec, gcrs6_2.dec, rtol=1e-8)

    # now try explicit intermediate pathways and ensure they're all consistent
    gcrs3 = (
        gcrs.transform_to(ITRS())
        .transform_to(CIRS())
        .transform_to(ITRS())
        .transform_to(gcrs)
    )
    assert_allclose(gcrs.ra, gcrs3.ra)
    assert_allclose(gcrs.dec, gcrs3.dec)

    gcrs4 = (
        gcrs.transform_to(ICRS())
        .transform_to(CIRS())
        .transform_to(ICRS())
        .transform_to(gcrs)
    )
    assert_allclose(gcrs.ra, gcrs4.ra)
    assert_allclose(gcrs.dec, gcrs4.dec)


def test_gcrs_altaz():
    """
    Check GCRS<->AltAz transforms for round-tripping.  Has multiple paths
    """
    from astropy.coordinates import EarthLocation

    usph = golden_spiral_grid(128)
    gcrs = GCRS(usph, obstime="J2000")[None]  # broadcast with times below

    # check array times sure N-d arrays work
    times = Time(np.linspace(2456293.25, 2456657.25, 51) * u.day, format="jd")[:, None]

    loc = EarthLocation(lon=10 * u.deg, lat=80.0 * u.deg)
    aaframe = AltAz(obstime=times, location=loc)

    aa1 = gcrs.transform_to(aaframe)
    aa2 = gcrs.transform_to(ICRS()).transform_to(CIRS()).transform_to(aaframe)
    aa3 = gcrs.transform_to(ITRS()).transform_to(CIRS()).transform_to(aaframe)

    # make sure they're all consistent
    assert_allclose(aa1.alt, aa2.alt)
    assert_allclose(aa1.az, aa2.az)
    assert_allclose(aa1.alt, aa3.alt)
    assert_allclose(aa1.az, aa3.az)


def test_gcrs_hadec():
    """
    Check GCRS<->HADec transforms for round-tripping.  Has multiple paths
    """
    from astropy.coordinates import EarthLocation

    usph = golden_spiral_grid(128)
    gcrs = GCRS(usph, obstime="J2000")  # broadcast with times below

    # check array times sure N-d arrays work
    times = Time(np.linspace(2456293.25, 2456657.25, 51) * u.day, format="jd")[:, None]

    loc = EarthLocation(lon=10 * u.deg, lat=80.0 * u.deg)
    hdframe = HADec(obstime=times, location=loc)

    hd1 = gcrs.transform_to(hdframe)
    hd2 = gcrs.transform_to(ICRS()).transform_to(CIRS()).transform_to(hdframe)
    hd3 = gcrs.transform_to(ITRS()).transform_to(CIRS()).transform_to(hdframe)

    # make sure they're all consistent
    assert_allclose(hd1.dec, hd2.dec)
    assert_allclose(hd1.ha, hd2.ha)
    assert_allclose(hd1.dec, hd3.dec)
    assert_allclose(hd1.ha, hd3.ha)


def test_precessed_geocentric():
    assert PrecessedGeocentric().equinox.jd == Time("J2000").jd

    gcrs_coo = GCRS(180 * u.deg, 2 * u.deg, distance=10000 * u.km)
    pgeo_coo = gcrs_coo.transform_to(PrecessedGeocentric())
    assert np.abs(gcrs_coo.ra - pgeo_coo.ra) > 10 * u.marcsec
    assert np.abs(gcrs_coo.dec - pgeo_coo.dec) > 10 * u.marcsec
    assert_allclose(gcrs_coo.distance, pgeo_coo.distance)

    gcrs_roundtrip = pgeo_coo.transform_to(GCRS())
    assert_allclose(gcrs_coo.ra, gcrs_roundtrip.ra)
    assert_allclose(gcrs_coo.dec, gcrs_roundtrip.dec)
    assert_allclose(gcrs_coo.distance, gcrs_roundtrip.distance)

    pgeo_coo2 = gcrs_coo.transform_to(PrecessedGeocentric(equinox="B1850"))
    assert np.abs(gcrs_coo.ra - pgeo_coo2.ra) > 1.5 * u.deg
    assert np.abs(gcrs_coo.dec - pgeo_coo2.dec) > 0.5 * u.deg
    assert_allclose(gcrs_coo.distance, pgeo_coo2.distance)

    gcrs2_roundtrip = pgeo_coo2.transform_to(GCRS())
    assert_allclose(gcrs_coo.ra, gcrs2_roundtrip.ra)
    assert_allclose(gcrs_coo.dec, gcrs2_roundtrip.dec)
    assert_allclose(gcrs_coo.distance, gcrs2_roundtrip.distance)


def test_precessed_geocentric_different_obstime():
    # Create two PrecessedGeocentric frames with different obstime
    precessedgeo1 = PrecessedGeocentric(obstime="2021-09-07")
    precessedgeo2 = PrecessedGeocentric(obstime="2021-06-07")

    # GCRS->PrecessedGeocentric should give different results for the two frames
    gcrs_coord = GCRS(10 * u.deg, 20 * u.deg, 3 * u.AU, obstime=precessedgeo1.obstime)
    pg_coord1 = gcrs_coord.transform_to(precessedgeo1)
    pg_coord2 = gcrs_coord.transform_to(precessedgeo2)
    assert not pg_coord1.is_equivalent_frame(pg_coord2)
    assert not allclose(pg_coord1.cartesian.xyz, pg_coord2.cartesian.xyz)

    # Looping back to GCRS should return the original coordinate
    loopback1 = pg_coord1.transform_to(gcrs_coord)
    loopback2 = pg_coord2.transform_to(gcrs_coord)
    assert loopback1.is_equivalent_frame(gcrs_coord)
    assert loopback2.is_equivalent_frame(gcrs_coord)
    assert_allclose(loopback1.cartesian.xyz, gcrs_coord.cartesian.xyz)
    assert_allclose(loopback2.cartesian.xyz, gcrs_coord.cartesian.xyz)


# shared by parametrized tests below.  Some use the whole AltAz, others use just obstime
totest_frames = [
    # J2000 is often a default so this might work when others don't
    AltAz(location=EarthLocation(-90 * u.deg, 65 * u.deg), obstime=Time("J2000")),
    AltAz(location=EarthLocation(120 * u.deg, -35 * u.deg), obstime=Time("J2000")),
    AltAz(
        location=EarthLocation(-90 * u.deg, 65 * u.deg),
        obstime=Time("2014-01-01 00:00:00"),
    ),
    AltAz(
        location=EarthLocation(-90 * u.deg, 65 * u.deg),
        obstime=Time("2014-08-01 08:00:00"),
    ),
    AltAz(
        location=EarthLocation(120 * u.deg, -35 * u.deg),
        obstime=Time("2014-01-01 00:00:00"),
    ),
]
MOONDIST = 385000 * u.km  # approximate moon semi-major orbit axis of moon
MOONDIST_CART = CartesianRepresentation(
    3**-0.5 * MOONDIST, 3**-0.5 * MOONDIST, 3**-0.5 * MOONDIST
)
# roughly earth orbital eccentricity, but with an added tolerance
EARTHECC = 0.017 + 0.005


@pytest.mark.parametrize("testframe", totest_frames)
def test_gcrs_altaz_sunish(testframe):
    """
    Sanity-check that the sun is at a reasonable distance from any altaz
    """
    sun = get_sun(testframe.obstime)

    assert sun.frame.name == "gcrs"

    # the .to(u.au) is not necessary, it just makes the asserts on failure more readable
    assert (EARTHECC - 1) * u.au < sun.distance.to(u.au) < (EARTHECC + 1) * u.au

    sunaa = sun.transform_to(testframe)
    assert (EARTHECC - 1) * u.au < sunaa.distance.to(u.au) < (EARTHECC + 1) * u.au


@pytest.mark.parametrize("testframe", totest_frames)
def test_gcrs_altaz_moonish(testframe):
    """
    Sanity-check that an object resembling the moon goes to the right place with
    a GCRS->AltAz transformation
    """
    moon = GCRS(MOONDIST_CART, obstime=testframe.obstime)

    moonaa = moon.transform_to(testframe)

    # now check that the distance change is similar to earth radius
    assert 1000 * u.km < np.abs(moonaa.distance - moon.distance).to(u.au) < 7000 * u.km

    # now check that it round-trips
    moon2 = moonaa.transform_to(moon)
    assert_allclose(moon.cartesian.xyz, moon2.cartesian.xyz)

    # also should add checks that the alt/az are different for different earth locations


@pytest.mark.parametrize("testframe", totest_frames)
def test_gcrs_altaz_bothroutes(testframe):
    """
    Repeat of both the moonish and sunish tests above to make sure the two
    routes through the coordinate graph are consistent with each other
    """
    sun = get_sun(testframe.obstime)
    sunaa_viaicrs = sun.transform_to(ICRS()).transform_to(testframe)
    sunaa_viaitrs = sun.transform_to(ITRS(obstime=testframe.obstime)).transform_to(
        testframe
    )

    moon = GCRS(MOONDIST_CART, obstime=testframe.obstime)
    moonaa_viaicrs = moon.transform_to(ICRS()).transform_to(testframe)
    moonaa_viaitrs = moon.transform_to(ITRS(obstime=testframe.obstime)).transform_to(
        testframe
    )

    assert_allclose(sunaa_viaicrs.cartesian.xyz, sunaa_viaitrs.cartesian.xyz)
    assert_allclose(moonaa_viaicrs.cartesian.xyz, moonaa_viaitrs.cartesian.xyz)


@pytest.mark.parametrize("testframe", totest_frames)
def test_cirs_altaz_moonish(testframe):
    """
    Sanity-check that an object resembling the moon goes to the right place with
    a CIRS<->AltAz transformation
    """
    moon = CIRS(MOONDIST_CART, obstime=testframe.obstime)

    moonaa = moon.transform_to(testframe)
    assert 1000 * u.km < np.abs(moonaa.distance - moon.distance).to(u.km) < 7000 * u.km

    # now check that it round-trips
    moon2 = moonaa.transform_to(moon)
    assert_allclose(moon.cartesian.xyz, moon2.cartesian.xyz)


@pytest.mark.parametrize("testframe", totest_frames)
def test_cirs_altaz_nodist(testframe):
    """
    Check that a UnitSphericalRepresentation coordinate round-trips for the
    CIRS<->AltAz transformation.
    """
    coo0 = CIRS(
        UnitSphericalRepresentation(10 * u.deg, 20 * u.deg), obstime=testframe.obstime
    )

    # check that it round-trips
    coo1 = coo0.transform_to(testframe).transform_to(coo0)
    assert_allclose(coo0.cartesian.xyz, coo1.cartesian.xyz)


@pytest.mark.parametrize("testframe", totest_frames)
def test_cirs_icrs_moonish(testframe):
    """
    check that something like the moon goes to about the right distance from the
    ICRS origin when starting from CIRS
    """
    moonish = CIRS(MOONDIST_CART, obstime=testframe.obstime)
    moonicrs = moonish.transform_to(ICRS())

    assert 0.97 * u.au < moonicrs.distance < 1.03 * u.au


@pytest.mark.parametrize("testframe", totest_frames)
def test_gcrs_icrs_moonish(testframe):
    """
    check that something like the moon goes to about the right distance from the
    ICRS origin when starting from GCRS
    """
    moonish = GCRS(MOONDIST_CART, obstime=testframe.obstime)
    moonicrs = moonish.transform_to(ICRS())

    assert 0.97 * u.au < moonicrs.distance < 1.03 * u.au


@pytest.mark.parametrize("testframe", totest_frames)
def test_icrs_gcrscirs_sunish(testframe):
    """
    check that the ICRS barycenter goes to about the right distance from various
    ~geocentric frames (other than testframe)
    """
    # slight offset to avoid divide-by-zero errors
    icrs = ICRS(0 * u.deg, 0 * u.deg, distance=10 * u.km)

    gcrs = icrs.transform_to(GCRS(obstime=testframe.obstime))
    assert (EARTHECC - 1) * u.au < gcrs.distance.to(u.au) < (EARTHECC + 1) * u.au

    cirs = icrs.transform_to(CIRS(obstime=testframe.obstime))
    assert (EARTHECC - 1) * u.au < cirs.distance.to(u.au) < (EARTHECC + 1) * u.au

    itrs = icrs.transform_to(ITRS(obstime=testframe.obstime))
    assert (
        (EARTHECC - 1) * u.au < itrs.spherical.distance.to(u.au) < (EARTHECC + 1) * u.au
    )


@pytest.mark.parametrize("testframe", totest_frames)
def test_icrs_altaz_moonish(testframe):
    """
    Check that something expressed in *ICRS* as being moon-like goes to the
    right AltAz distance
    """
    # we use epv00 instead of get_sun because get_sun includes aberration
    earth_pv_helio, earth_pv_bary = erfa.epv00(*get_jd12(testframe.obstime, "tdb"))
    earth_icrs_xyz = earth_pv_bary[0] * u.au
    moonoffset = [0, 0, MOONDIST.value] * MOONDIST.unit
    moonish_icrs = ICRS(CartesianRepresentation(earth_icrs_xyz + moonoffset))
    moonaa = moonish_icrs.transform_to(testframe)

    # now check that the distance change is similar to earth radius
    assert 1000 * u.km < np.abs(moonaa.distance - MOONDIST).to(u.au) < 7000 * u.km


def test_gcrs_self_transform_closeby():
    """
    Tests GCRS self transform for objects which are nearby and thus
    have reasonable parallax.

    Moon positions were originally created using JPL DE432s ephemeris.

    The two lunar positions (one geocentric, one at a defined location)
    are created via a transformation from ICRS to two different GCRS frames.

    We test that the GCRS-GCRS self transform can correctly map one GCRS
    frame onto the other.
    """
    t = Time("2014-12-25T07:00")
    moon_geocentric = SkyCoord(
        GCRS(
            318.10579159 * u.deg,
            -11.65281165 * u.deg,
            365042.64880308 * u.km,
            obstime=t,
        )
    )

    # this is the location of the Moon as seen from La Palma
    obsgeoloc = [-5592982.59658935, -63054.1948592, 3059763.90102216] * u.m
    obsgeovel = [4.59798494, -407.84677071, 0.0] * u.m / u.s
    moon_lapalma = SkyCoord(
        GCRS(
            318.7048445 * u.deg,
            -11.98761996 * u.deg,
            369722.8231031 * u.km,
            obstime=t,
            obsgeoloc=obsgeoloc,
            obsgeovel=obsgeovel,
        )
    )

    transformed = moon_geocentric.transform_to(moon_lapalma.frame)
    delta = transformed.separation_3d(moon_lapalma)
    assert_allclose(delta, 0.0 * u.m, atol=1 * u.m)


def test_teme_itrf():
    """
    Test case transform from TEME to ITRF.

    Test case derives from example on appendix C of Vallado, Crawford, Hujsak & Kelso (2006).
    See https://celestrak.com/publications/AIAA/2006-6753/AIAA-2006-6753-Rev2.pdf
    """
    v_itrf = CartesianDifferential(
        -3.225636520, -2.872451450, 5.531924446, unit=u.km / u.s
    )
    p_itrf = CartesianRepresentation(
        -1033.479383,
        7901.2952740,
        6380.35659580,
        unit=u.km,
        differentials={"s": v_itrf},
    )
    t = Time("2004-04-06T07:51:28.386")

    teme = ITRS(p_itrf, obstime=t).transform_to(TEME(obstime=t))
    v_teme = CartesianDifferential(
        -4.746131487, 0.785818041, 5.531931288, unit=u.km / u.s
    )
    p_teme = CartesianRepresentation(
        5094.18016210,
        6127.64465050,
        6380.34453270,
        unit=u.km,
        differentials={"s": v_teme},
    )

    assert_allclose(
        teme.cartesian.without_differentials().xyz,
        p_teme.without_differentials().xyz,
        atol=30 * u.cm,
    )

    assert_allclose(
        teme.cartesian.differentials["s"].d_xyz,
        p_teme.differentials["s"].d_xyz,
        atol=1.0 * u.cm / u.s,
    )

    # test round trip
    itrf = teme.transform_to(ITRS(obstime=t))
    assert_allclose(
        itrf.cartesian.without_differentials().xyz,
        p_itrf.without_differentials().xyz,
        atol=100 * u.cm,
    )
    assert_allclose(
        itrf.cartesian.differentials["s"].d_xyz,
        p_itrf.differentials["s"].d_xyz,
        atol=1 * u.cm / u.s,
    )


def test_precessedgeocentric_loopback():
    from_coo = PrecessedGeocentric(
        1 * u.deg, 2 * u.deg, 3 * u.AU, obstime="2001-01-01", equinox="2001-01-01"
    )

    # Change just the obstime
    to_frame = PrecessedGeocentric(obstime="2001-06-30", equinox="2001-01-01")

    explicit_coo = from_coo.transform_to(ICRS()).transform_to(to_frame)
    implicit_coo = from_coo.transform_to(to_frame)

    # Confirm that the explicit transformation changes the coordinate
    assert not allclose(explicit_coo.ra, from_coo.ra, rtol=1e-10)
    assert not allclose(explicit_coo.dec, from_coo.dec, rtol=1e-10)
    assert not allclose(explicit_coo.distance, from_coo.distance, rtol=1e-10)

    # Confirm that the loopback matches the explicit transformation
    assert_allclose(explicit_coo.ra, implicit_coo.ra, rtol=1e-10)
    assert_allclose(explicit_coo.dec, implicit_coo.dec, rtol=1e-10)
    assert_allclose(explicit_coo.distance, implicit_coo.distance, rtol=1e-10)

    # Change just the equinox
    to_frame = PrecessedGeocentric(obstime="2001-01-01", equinox="2001-06-30")

    explicit_coo = from_coo.transform_to(ICRS()).transform_to(to_frame)
    implicit_coo = from_coo.transform_to(to_frame)

    # Confirm that the explicit transformation changes the direction but not the distance
    assert not allclose(explicit_coo.ra, from_coo.ra, rtol=1e-10)
    assert not allclose(explicit_coo.dec, from_coo.dec, rtol=1e-10)
    assert allclose(explicit_coo.distance, from_coo.distance, rtol=1e-10)

    # Confirm that the loopback matches the explicit transformation
    assert_allclose(explicit_coo.ra, implicit_coo.ra, rtol=1e-10)
    assert_allclose(explicit_coo.dec, implicit_coo.dec, rtol=1e-10)
    assert_allclose(explicit_coo.distance, implicit_coo.distance, rtol=1e-10)


def test_teme_loopback():
    from_coo = TEME(1 * u.AU, 2 * u.AU, 3 * u.AU, obstime="2001-01-01")
    to_frame = TEME(obstime="2001-06-30")

    explicit_coo = from_coo.transform_to(ICRS()).transform_to(to_frame)
    implicit_coo = from_coo.transform_to(to_frame)

    # Confirm that the explicit transformation changes the coordinate
    assert not allclose(explicit_coo.cartesian.xyz, from_coo.cartesian.xyz, rtol=1e-10)

    # Confirm that the loopback matches the explicit transformation
    assert_allclose(explicit_coo.cartesian.xyz, implicit_coo.cartesian.xyz, rtol=1e-10)


@pytest.mark.remote_data
def test_earth_orientation_table(monkeypatch):
    """Check that we can set the IERS table used as Earth Reference.

    Use the here and now to be sure we get a difference.
    """
    monkeypatch.setattr("astropy.utils.iers.conf.auto_download", True)
    t = Time.now()
    location = EarthLocation(lat=0 * u.deg, lon=0 * u.deg)
    altaz = AltAz(location=location, obstime=t)
    sc = SkyCoord(1 * u.deg, 2 * u.deg)
    # Default: uses IERS_Auto, which will give a prediction.
    # Note: tests run with warnings turned into errors, so it is
    # meaningful if this passes.
    if CI:
        with warnings.catch_warnings():
            # Server occasionally blocks IERS download in CI.
            warnings.filterwarnings("ignore", message=r".*using local IERS-B.*")
            # This also captures unclosed socket warning that is ignored in setup.cfg
            warnings.filterwarnings("ignore", message=r".*unclosed.*")
            altaz_auto = sc.transform_to(altaz)
    else:
        altaz_auto = sc.transform_to(altaz)  # No warnings

    with iers.earth_orientation_table.set(iers.IERS_B.open()):
        with pytest.warns(AstropyWarning, match="after IERS data"):
            altaz_b = sc.transform_to(altaz)

    sep_b_auto = altaz_b.separation(altaz_auto)
    assert_allclose(sep_b_auto, 0.0 * u.deg, atol=1 * u.arcsec)
    assert sep_b_auto > 10 * u.microarcsecond

    # Check we returned to regular IERS system.
    altaz_auto2 = sc.transform_to(altaz)
    assert_allclose(altaz_auto2.separation(altaz_auto), 0 * u.deg)


@pytest.mark.remote_data
@pytest.mark.skipif(not HAS_JPLEPHEM, reason="requires jplephem")
def test_ephemerides():
    """
    We test that using different ephemerides gives very similar results
    for transformations
    """
    t = Time("2014-12-25T07:00")
    moon = SkyCoord(
        GCRS(
            318.10579159 * u.deg,
            -11.65281165 * u.deg,
            365042.64880308 * u.km,
            obstime=t,
        )
    )

    icrs_frame = ICRS()
    hcrs_frame = HCRS(obstime=t)
    ecl_frame = HeliocentricMeanEcliptic(equinox=t)
    cirs_frame = CIRS(obstime=t)

    moon_icrs_builtin = moon.transform_to(icrs_frame)
    moon_hcrs_builtin = moon.transform_to(hcrs_frame)
    moon_helioecl_builtin = moon.transform_to(ecl_frame)
    moon_cirs_builtin = moon.transform_to(cirs_frame)

    with solar_system_ephemeris.set("jpl"):
        moon_icrs_jpl = moon.transform_to(icrs_frame)
        moon_hcrs_jpl = moon.transform_to(hcrs_frame)
        moon_helioecl_jpl = moon.transform_to(ecl_frame)
        moon_cirs_jpl = moon.transform_to(cirs_frame)

    # most transformations should differ by an amount which is
    # non-zero but of order milliarcsecs
    sep_icrs = moon_icrs_builtin.separation(moon_icrs_jpl)
    sep_hcrs = moon_hcrs_builtin.separation(moon_hcrs_jpl)
    sep_helioecl = moon_helioecl_builtin.separation(moon_helioecl_jpl)
    sep_cirs = moon_cirs_builtin.separation(moon_cirs_jpl)

    assert_allclose([sep_icrs, sep_hcrs, sep_helioecl], 0.0 * u.deg, atol=10 * u.mas)
    assert all(
        sep > 10 * u.microarcsecond for sep in (sep_icrs, sep_hcrs, sep_helioecl)
    )

    # CIRS should be the same
    assert_allclose(sep_cirs, 0.0 * u.deg, atol=1 * u.microarcsecond)


def test_tete_transforms():
    """
    We test the TETE transforms for proper behaviour here.

    The TETE transforms are tested for accuracy against JPL Horizons in
    test_solar_system.py. Here we are looking to check for consistency and
    errors in the self transform.
    """
    loc = EarthLocation.from_geodetic("-22°57'35.1", "-67°47'14.1", 5186 * u.m)
    time = Time("2020-04-06T00:00")
    p, v = loc.get_gcrs_posvel(time)

    gcrs_frame = GCRS(obstime=time, obsgeoloc=p, obsgeovel=v)
    moon = SkyCoord(
        169.24113968 * u.deg,
        10.86086666 * u.deg,
        358549.25381755 * u.km,
        frame=gcrs_frame,
    )

    tete_frame = TETE(obstime=time, location=loc)
    # need to set obsgeoloc/vel explicitly or skycoord behaviour over-writes
    tete_geo = TETE(obstime=time, location=EarthLocation(*([0, 0, 0] * u.km)))

    # test self-transform by comparing to GCRS-TETE-ITRS-TETE route
    tete_coo1 = moon.transform_to(tete_frame)
    tete_coo2 = moon.transform_to(tete_geo)
    assert_allclose(tete_coo1.separation_3d(tete_coo2), 0 * u.mm, atol=1 * u.mm)

    # test TETE-ITRS transform by comparing GCRS-CIRS-ITRS to GCRS-TETE-ITRS
    itrs1 = moon.transform_to(CIRS()).transform_to(ITRS())
    itrs2 = moon.transform_to(TETE()).transform_to(ITRS())
    assert_allclose(itrs1.separation_3d(itrs2), 0 * u.mm, atol=1 * u.mm)

    # test round trip GCRS->TETE->GCRS
    new_moon = moon.transform_to(TETE()).transform_to(moon)
    assert_allclose(new_moon.separation_3d(moon), 0 * u.mm, atol=1 * u.mm)

    # test round trip via ITRS
    tete_rt = tete_coo1.transform_to(ITRS(obstime=time)).transform_to(tete_coo1)
    assert_allclose(tete_rt.separation_3d(tete_coo1), 0 * u.mm, atol=1 * u.mm)

    # ensure deprecated routine remains consistent
    # make sure test raises warning!
    with pytest.warns(AstropyDeprecationWarning, match="The use of"):
        tete_alt = _apparent_position_in_true_coordinates(moon)
    assert_allclose(tete_coo1.separation_3d(tete_alt), 0 * u.mm, atol=100 * u.mm)


def test_straight_overhead():
    """
    With a precise CIRS<->Observed transformation this should give Alt=90 exactly

    If the CIRS self-transform breaks it won't, due to improper treatment of aberration
    """
    t = Time("J2010")
    obj = EarthLocation(-1 * u.deg, 52 * u.deg, height=10.0 * u.km)
    home = EarthLocation(-1 * u.deg, 52 * u.deg, height=0.0 * u.km)

    # An object that appears straight overhead - FOR A GEOCENTRIC OBSERVER.
    # Note, this won't be overhead for a topocentric observer because of
    # aberration.
    cirs_geo = obj.get_itrs(t).transform_to(CIRS(obstime=t))

    # now get the Geocentric CIRS position of observatory
    obsrepr = home.get_itrs(t).transform_to(CIRS(obstime=t)).cartesian

    # topocentric CIRS position of a straight overhead object
    cirs_repr = cirs_geo.cartesian - obsrepr

    # create a CIRS object that appears straight overhead for a TOPOCENTRIC OBSERVER
    topocentric_cirs_frame = CIRS(obstime=t, location=home)
    cirs_topo = topocentric_cirs_frame.realize_frame(cirs_repr)

    # Check AltAz (though Azimuth can be anything so is not tested).
    aa = cirs_topo.transform_to(AltAz(obstime=t, location=home))
    assert_allclose(aa.alt, 90 * u.deg, atol=1 * u.uas, rtol=0)

    # Check HADec.
    hd = cirs_topo.transform_to(HADec(obstime=t, location=home))
    assert_allclose(hd.ha, 0 * u.hourangle, atol=1 * u.uas, rtol=0)
    assert_allclose(hd.dec, 52 * u.deg, atol=1 * u.uas, rtol=0)


def test_itrs_straight_overhead():
    """
    With a precise ITRS<->Observed transformation this should give Alt=90 exactly

    """
    t = Time("J2010")
    obj = EarthLocation(-1 * u.deg, 52 * u.deg, height=10.0 * u.km)
    home = EarthLocation(-1 * u.deg, 52 * u.deg, height=0.0 * u.km)

    # Check AltAz (though Azimuth can be anything so is not tested).
    aa = obj.get_itrs(t, location=home).transform_to(AltAz(obstime=t, location=home))
    assert_allclose(aa.alt, 90 * u.deg, atol=1 * u.uas, rtol=0)

    # Check HADec.
    hd = obj.get_itrs(t, location=home).transform_to(HADec(obstime=t, location=home))
    assert_allclose(hd.ha, 0 * u.hourangle, atol=1 * u.uas, rtol=0)
    assert_allclose(hd.dec, 52 * u.deg, atol=1 * u.uas, rtol=0)


def jplephem_ge(minversion):
    """Check if jplephem is installed and has version >= minversion."""
    # This is a separate routine since somehow with pyinstaller the stanza
    # not HAS_JPLEPHEM or metadata.version('jplephem') < '2.15'
    # leads to a module not found error.
    try:
        return HAS_JPLEPHEM and metadata.version("jplephem") >= minversion
    except Exception:
        return False


@pytest.mark.remote_data
@pytest.mark.skipif(not jplephem_ge("2.15"), reason="requires jplephem >= 2.15")
def test_aa_hd_high_precision():
    """These tests are provided by @mkbrewer - see issue #10356.

    The code that produces them agrees very well (<0.5 mas) with SkyField once Polar motion
    is turned off, but SkyField does not include polar motion, so a comparison to Skyfield
    or JPL Horizons will be ~1" off.

    The absence of polar motion within Skyfield and the disagreement between Skyfield and Horizons
    make high precision comparisons to those codes difficult.

    Updated 2020-11-29, after the comparison between codes became even better,
    down to 100 nas.

    Updated 2023-02-14, after IERS changes the IERS B format and analysis,
    causing small deviations.

    NOTE: the agreement reflects consistency in approach between two codes,
    not necessarily absolute precision.  If this test starts failing, the
    tolerance can and should be weakened *if* it is clear that the change is
    due to an improvement (e.g., a new IAU precession model).

    """
    lat = -22.959748 * u.deg
    lon = -67.787260 * u.deg
    elev = 5186 * u.m
    loc = EarthLocation.from_geodetic(lon, lat, elev)
    # Note: at this level of precision for the comparison, we have to include
    # the location in the time, as it influences the transformation to TDB.
    t = Time("2017-04-06T00:00:00.0", location=loc)
    with solar_system_ephemeris.set("de430"):
        moon = get_body("moon", t, loc)
        moon_aa = moon.transform_to(AltAz(obstime=t, location=loc))
        moon_hd = moon.transform_to(HADec(obstime=t, location=loc))

    # Numbers from
    # https://github.com/astropy/astropy/pull/11073#issuecomment-735486271
    # updated in https://github.com/astropy/astropy/issues/11683
    # and again after the IERS_B change.
    TARGET_AZ, TARGET_EL = 15.032673662647138 * u.deg, 50.303110087520054 * u.deg
    TARGET_DISTANCE = 376252.88325051306 * u.km
    assert_allclose(moon_aa.az, TARGET_AZ, atol=0.1 * u.uas, rtol=0)
    assert_allclose(moon_aa.alt, TARGET_EL, atol=0.1 * u.uas, rtol=0)
    assert_allclose(moon_aa.distance, TARGET_DISTANCE, atol=0.1 * u.mm, rtol=0)
    ha, dec = erfa.ae2hd(
        moon_aa.az.to_value(u.radian),
        moon_aa.alt.to_value(u.radian),
        lat.to_value(u.radian),
    )
    ha = u.Quantity(ha, u.radian, copy=False)
    dec = u.Quantity(dec, u.radian, copy=False)
    assert_allclose(moon_hd.ha, ha, atol=0.1 * u.uas, rtol=0)
    assert_allclose(moon_hd.dec, dec, atol=0.1 * u.uas, rtol=0)


def test_aa_high_precision_nodata():
    """
    These tests are designed to ensure high precision alt-az transforms.

    They are a slight fudge since the target values come from astropy itself. They are generated
    with a version of the code that passes the tests above, but for the internal solar system
    ephemerides to avoid the use of remote data.
    """
    # Last updated when the new IERS B format and analysis was introduced.
    TARGET_AZ, TARGET_EL = 15.0323151 * u.deg, 50.30271925 * u.deg
    lat = -22.959748 * u.deg
    lon = -67.787260 * u.deg
    elev = 5186 * u.m
    loc = EarthLocation.from_geodetic(lon, lat, elev)
    t = Time("2017-04-06T00:00:00.0")

    moon = get_body("moon", t, loc)
    moon_aa = moon.transform_to(AltAz(obstime=t, location=loc))
    assert_allclose(moon_aa.az - TARGET_AZ, 0 * u.mas, atol=0.5 * u.mas)
    assert_allclose(moon_aa.alt - TARGET_EL, 0 * u.mas, atol=0.5 * u.mas)


class TestGetLocationGCRS:
    # TETE and CIRS use get_location_gcrs to get obsgeoloc and obsgeovel
    # with knowledge of some of the matrices. Check that this is consistent
    # with a direct transformation.
    def setup_class(cls):
        cls.loc = loc = EarthLocation.from_geodetic(
            np.linspace(0, 360, 6) * u.deg, np.linspace(-90, 90, 6) * u.deg, 100 * u.m
        )
        cls.obstime = obstime = Time(np.linspace(2000, 2010, 6), format="jyear")
        # Get comparison via a full transformation.  We do not use any methods
        # of EarthLocation, since those depend on the fast transform.
        loc_itrs = ITRS(loc.x, loc.y, loc.z, obstime=obstime)
        zeros = np.broadcast_to(0.0 * (u.km / u.s), (3,) + loc_itrs.shape, subok=True)
        loc_itrs.data.differentials["s"] = CartesianDifferential(zeros)
        loc_gcrs_cart = loc_itrs.transform_to(GCRS(obstime=obstime)).cartesian
        cls.obsgeoloc = loc_gcrs_cart.without_differentials()
        cls.obsgeovel = loc_gcrs_cart.differentials["s"].to_cartesian()

    def check_obsgeo(self, obsgeoloc, obsgeovel):
        assert_allclose(obsgeoloc.xyz, self.obsgeoloc.xyz, atol=0.1 * u.um, rtol=0.0)
        assert_allclose(
            obsgeovel.xyz, self.obsgeovel.xyz, atol=0.1 * u.mm / u.s, rtol=0.0
        )

    def test_get_gcrs_posvel(self):
        # Really just a sanity check
        self.check_obsgeo(*self.loc.get_gcrs_posvel(self.obstime))

    def test_tete_quick(self):
        # Following copied from intermediate_rotation_transforms.gcrs_to_tete
        rbpn = erfa.pnm06a(*get_jd12(self.obstime, "tt"))
        loc_gcrs_frame = get_location_gcrs(
            self.loc, self.obstime, tete_to_itrs_mat(self.obstime, rbpn=rbpn), rbpn
        )
        self.check_obsgeo(loc_gcrs_frame.obsgeoloc, loc_gcrs_frame.obsgeovel)

    def test_cirs_quick(self):
        cirs_frame = CIRS(location=self.loc, obstime=self.obstime)
        # Following copied from intermediate_rotation_transforms.gcrs_to_cirs
        pmat = gcrs_to_cirs_mat(cirs_frame.obstime)
        loc_gcrs_frame = get_location_gcrs(
            self.loc, self.obstime, cirs_to_itrs_mat(cirs_frame.obstime), pmat
        )
        self.check_obsgeo(loc_gcrs_frame.obsgeoloc, loc_gcrs_frame.obsgeovel)



# Licensed under a 3-clause BSD style license - see LICENSE.rst

import itertools

import numpy as np
import pytest
from numpy.testing import assert_allclose, assert_almost_equal

from astropy import units as u
from astropy.convolution.convolve import convolve, convolve_fft
from astropy.convolution.kernels import (
    Box2DKernel,
    Gaussian2DKernel,
    Moffat2DKernel,
    Tophat2DKernel,
)

SHAPES_ODD = [[15, 15], [31, 31]]
SHAPES_EVEN = [[8, 8], [16, 16], [32, 32]]  # FIXME: not used ?!
NOSHAPE = [[None, None]]
WIDTHS = [2, 3, 4, 5]

KERNELS = []

for shape in SHAPES_ODD + NOSHAPE:
    for width in WIDTHS:
        KERNELS.append(
            Gaussian2DKernel(
                x_stddev=width,
                x_size=shape[0],
                y_size=shape[1],
                mode="oversample",
                factor=10,
            )
        )

        KERNELS.append(
            Box2DKernel(
                width=width,
                x_size=shape[0],
                y_size=shape[1],
                mode="oversample",
                factor=10,
            )
        )

        KERNELS.append(
            Tophat2DKernel(
                radius=width,
                x_size=shape[0],
                y_size=shape[1],
                mode="oversample",
                factor=10,
            )
        )
        KERNELS.append(
            Moffat2DKernel(
                gamma=width,
                alpha=2,
                x_size=shape[0],
                y_size=shape[1],
                mode="oversample",
                factor=10,
            )
        )


class Test2DConvolutions:
    @pytest.mark.parametrize("kernel", KERNELS)
    def test_centered_makekernel(self, kernel):
        """
        Test smoothing of an image with a single positive pixel
        """

        shape = kernel.array.shape

        x = np.zeros(shape)
        xslice = tuple(slice(sh // 2, sh // 2 + 1) for sh in shape)
        x[xslice] = 1.0

        c2 = convolve_fft(x, kernel, boundary="fill")
        c1 = convolve(x, kernel, boundary="fill")

        assert_almost_equal(c1, c2, decimal=12)

    @pytest.mark.parametrize("kernel", KERNELS)
    def test_random_makekernel(self, kernel):
        """
        Test smoothing of an image made of random noise
        """

        shape = kernel.array.shape

        x = np.random.randn(*shape)

        c2 = convolve_fft(x, kernel, boundary="fill")
        c1 = convolve(x, kernel, boundary="fill")

        # not clear why, but these differ by a couple ulps...
        assert_almost_equal(c1, c2, decimal=12)

    @pytest.mark.parametrize(
        ("shape", "width"), list(itertools.product(SHAPES_ODD, WIDTHS))
    )
    def test_uniform_smallkernel(self, shape, width):
        """
        Test smoothing of an image with a single positive pixel

        Uses a simple, small kernel
        """

        if width % 2 == 0:
            # convolve does not accept odd-shape kernels
            return

        kernel = np.ones([width, width])

        x = np.zeros(shape)
        xslice = tuple(slice(sh // 2, sh // 2 + 1) for sh in shape)
        x[xslice] = 1.0

        c2 = convolve_fft(x, kernel, boundary="fill")
        c1 = convolve(x, kernel, boundary="fill")

        assert_almost_equal(c1, c2, decimal=12)

    @pytest.mark.parametrize(
        ("shape", "width"), list(itertools.product(SHAPES_ODD, [1, 3, 5]))
    )
    def test_smallkernel_Box2DKernel(self, shape, width):
        """
        Test smoothing of an image with a single positive pixel

        Compares a small uniform kernel to the Box2DKernel
        """

        kernel1 = np.ones([width, width]) / float(width) ** 2
        kernel2 = Box2DKernel(width, mode="oversample", factor=10)

        x = np.zeros(shape)
        xslice = tuple(slice(sh // 2, sh // 2 + 1) for sh in shape)
        x[xslice] = 1.0

        c2 = convolve_fft(x, kernel2, boundary="fill")
        c1 = convolve_fft(x, kernel1, boundary="fill")

        assert_almost_equal(c1, c2, decimal=12)

        c2 = convolve(x, kernel2, boundary="fill")
        c1 = convolve(x, kernel1, boundary="fill")

        assert_almost_equal(c1, c2, decimal=12)


def test_gaussian_2d_kernel_quantity():
    # Make sure that the angle can be a quantity
    kernel1 = Gaussian2DKernel(x_stddev=2, y_stddev=4, theta=45 * u.deg)
    kernel2 = Gaussian2DKernel(x_stddev=2, y_stddev=4, theta=np.pi / 4)
    assert_allclose(kernel1.array, kernel2.array)



# Licensed under a 3-clause BSD style license - see LICENSE.rst

# The BoxLeastSquares periodogram functionality has been moved to
# astropy.timeseries.periodograms.bls. The purpose of this file is to provide backward-
# compatibility during a transition phase. We can't emit a deprecation warning
# simply on import of this module, since the classes are imported into the
# top-level astropy.stats, so instead we wrap the main class and emit a
# warning during initialization.

import warnings

from astropy.timeseries.periodograms.bls import (
    BoxLeastSquares as TimeseriesBoxLeastSquares,
)
from astropy.timeseries.periodograms.bls import (
    BoxLeastSquaresResults as TimeseriesBoxLeastSquaresResults,
)
from astropy.utils.exceptions import AstropyDeprecationWarning

__all__ = ["BoxLeastSquares", "BoxLeastSquaresResults"]


class BoxLeastSquares(TimeseriesBoxLeastSquares):
    """
    Compute the box least squares periodogram.

    This class has been deprecated and will be removed in a future version.
    Use `astropy.timeseries.BoxLeastSquares` instead.
    """

    def __init__(self, *args, **kwargs):
        warnings.warn(
            "Importing BoxLeastSquares from astropy.stats has been "
            "deprecated and will no longer be supported in future. "
            "Please import this class from the astropy.timeseries "
            "module instead",
            AstropyDeprecationWarning,
        )
        super().__init__(*args, **kwargs)


class BoxLeastSquaresResults(TimeseriesBoxLeastSquaresResults):
    """
    The results of a BoxLeastSquares search.

    This class has been deprecated and will be removed in a future version.
    Use `astropy.timeseries.BoxLeastSquaresResults` instead.
    """

    def __init__(self, *args, **kwargs):
        warnings.warn(
            "Importing BoxLeastSquaresResults from astropy.stats has been "
            "deprecated and will no longer be supported in future. "
            "Please import this class from the astropy.timeseries "
            "module instead",
            AstropyDeprecationWarning,
        )
        super().__init__(*args, **kwargs)



# Licensed under a 3-clause BSD style license
"""
:Author: Simon Gibbons (simongibbons@gmail.com).
"""


from .core import DefaultSplitter
from .fixedwidth import (
    FixedWidth,
    FixedWidthData,
    FixedWidthHeader,
    FixedWidthTwoLineDataSplitter,
)


class SimpleRSTHeader(FixedWidthHeader):
    position_line = 0
    start_line = 1
    splitter_class = DefaultSplitter
    position_char = "="

    def get_fixedwidth_params(self, line):
        vals, starts, ends = super().get_fixedwidth_params(line)
        # The right hand column can be unbounded
        ends[-1] = None
        return vals, starts, ends


class SimpleRSTData(FixedWidthData):
    end_line = -1
    splitter_class = FixedWidthTwoLineDataSplitter


class RST(FixedWidth):
    """reStructuredText simple format table.

    See: https://docutils.sourceforge.io/docs/ref/rst/restructuredtext.html#simple-tables

    Example::

      >>> from astropy.table import QTable
      >>> import astropy.units as u
      >>> import sys
      >>> tbl = QTable({"wave": [350, 950] * u.nm, "response": [0.7, 1.2] * u.count})
      >>> tbl.write(sys.stdout,  format="ascii.rst")
      ===== ========
       wave response
      ===== ========
      350.0      0.7
      950.0      1.2
      ===== ========

    Like other fixed-width formats, when writing a table you can provide ``header_rows``
    to specify a list of table rows to output as the header.  For example::

      >>> tbl.write(sys.stdout,  format="ascii.rst", header_rows=['name', 'unit'])
      ===== ========
       wave response
         nm       ct
      ===== ========
      350.0      0.7
      950.0      1.2
      ===== ========

    Currently there is no support for reading tables which utilize continuation lines,
    or for ones which define column spans through the use of an additional
    line of dashes in the header.

    """

    _format_name = "rst"
    _description = "reStructuredText simple table"
    data_class = SimpleRSTData
    header_class = SimpleRSTHeader

    def __init__(self, header_rows=None):
        super().__init__(delimiter_pad=None, bookend=False, header_rows=header_rows)

    def write(self, lines):
        lines = super().write(lines)
        idx = len(self.header.header_rows)
        lines = [lines[idx]] + lines + [lines[idx]]
        return lines

    def read(self, table):
        self.data.start_line = 2 + len(self.header.header_rows)
        return super().read(table)
